"""This module contains functions to perform calculations for lock operations."""

import numpy as np
from opentnsim.utils import time_to_numpy
from opentnsim.constants import gravitational_acceleration
from opentnsim.vessel_traffic_service.hydrodanamic_data_manager import HydrodynamicDataManager


def calculate_z(
    t,
    t_start,
    direction,
    wlev_init,
    operation_index,
    operation_planning,
    hydrodynamic_information_path,
    start_node,
    end_node,
    node_open,
    epoch,
):
    # set default time and water level difference series
    z = np.zeros_like(t)

    # convert given t_start into np.datetime64 (this is required to communicate with the hydrodynamic data via the NetCDF package)
    t_start = time_to_numpy(t_start)

    hydromanager = HydrodynamicDataManager()
    # determine the actual water levels
    time_index = hydromanager._get_time_index_of_hydrodynamic_data(hydrodynamic_information_path, t_start)
    t_simulation_start = np.datetime64(epoch)
    H_A = hydromanager._get_hydrodynamic_data_series(hydrodynamic_information_path, t_simulation_start, start_node, "Water level")
    H_B = hydromanager._get_hydrodynamic_data_series(hydrodynamic_information_path, t_simulation_start, end_node, "Water level")
    H_A_init = H_A[time_index]
    H_B_init = H_B[time_index]

    if wlev_init is None:
        last_operations = operation_planning[operation_planning.index < operation_index]
        if not last_operations.empty:
            last_operation = last_operations.iloc[-1]
            last_operation_direction = last_operation.direction
            if not last_operation_direction:
                wlev_init = H_B_init
            else:
                wlev_init = H_A_init

        elif node_open == start_node:
            wlev_init = H_A_init
        else:
            wlev_init = H_B_init

    if not direction:
        z[0] = H_B_init - wlev_init

    else:
        z[0] = H_A_init - wlev_init

    return z, H_A, H_B


def levelling_time_equation(
    t,
    z,
    lock_length,
    lock_width,
    disch_coeff,
    gate_opening_time,
    opening_area,
    t_start,
    dt,
    direction,
    water_level_difference_limit_to_open_doors,
    prediction,
    H_A,
    H_B,
):
    """Calculates the levelling time of a lock operation based on Eq. 4.64 of Ports and Waterways Open Textbook (https://books.open.tudelft.nl/home/catalog/book/204)
    This function is called by determine_levelling_time()
    Returns
    -------
    levelling_time : float
        the time duration of the levelling process
    t : list of float
        the time series of the levelling process
    z : list of float
        the water level difference series over the time of the levelling process
    """
    t_start = time_to_numpy(t_start)
    A_ch = lock_length * lock_width  # surface area of the lock chamber [m^2] (constant over time)
    m = disch_coeff  # discharge coefficient [-] (constant over time)
    g = gravitational_acceleration  # gravitational acceleration [m/(s^2)] (constant over time)
    T1 = gate_opening_time  # time to open the gate [s] (constant over time)
    A_s = np.linspace(0, opening_area, int(T1 / float(dt)))  # sluice opening area over time when opening [m^2] (time-dependent)
    A_s = np.append(A_s, [opening_area] * (len(z) - len(A_s)))  # sluice opening over full levelling process [m^2] (time-dependent)
    H_time = HydrodynamicDataManager().hydrodynamic_times.astype(float)  # time series of the hydrodynamic data [s]

    # time-integration by (self-coded) Euler's method TODO Checken of we een standaard solver kunnen gebruiken. En of we dit algoritme los kunnen maken van de klasse.
    for i in range(len(t) - 1):
        H_Ai = np.interp(
            (np.timedelta64(int(i * float(dt) * 10**6), "us") + t_start - np.datetime64("1970-01-01")) / np.timedelta64(1, "us"),
            H_time,
            H_A,
        )  # water level at side A at time = i
        H_Aii = np.interp(
            (np.timedelta64(int((i + 1) * float(dt) * 10**6), "us") + t_start - np.datetime64("1970-01-01"))
            / np.timedelta64(1, "us"),
            H_time,
            H_A,
        )  # water level at side A at time = i + 1
        H_Bi = np.interp(
            (np.timedelta64(int(i * float(dt) * 10**6), "us") + t_start - np.datetime64("1970-01-01")) / np.timedelta64(1, "us"),
            H_time,
            H_B,
        )  # water level at side B at time = i
        H_Bii = np.interp(
            (np.timedelta64(int((i + 1) * float(dt) * 10**6), "us") + t_start - np.datetime64("1970-01-01"))
            / np.timedelta64(1, "us"),
            H_time,
            H_B,
        )  # water level at side B at time = i + 1
        deltaH_A = H_Aii - H_Ai  # water level difference at side A between time = i and time = i + 1
        deltaH_B = H_Bii - H_Bi  # water level difference at side B between time = i and time = i + 1

        # determine the contribution to the change in water level difference outside of the lock (i.e., due to tides) in the water level difference at time = i + 1
        if not direction:
            to_wlev_change = -deltaH_B
        else:
            to_wlev_change = -deltaH_A

        # calculate change in water level difference between time = i and time = i + 1
        z_i = abs(z[i])  # absolute water level difference at time = i

        dz_dt = -m * A_s[i] * np.sqrt(2 * g * np.max([0, z_i])) / A_ch  # change in water level difference over time [m/s]
        if z[i] < 0:  # correct if water level difference is negative
            dz_dt = -dz_dt
        dz = dz_dt * float(dt) + to_wlev_change

        # calculate the new water level difference at time = i + 1
        z[i + 1] = z[i] + dz
        if np.sign(z[i + 1]) != np.sign(z[i]):  # prevents overshooting of the water level difference
            z[i + 1] = 0

        if (
            np.abs(z[i + 1]) <= water_level_difference_limit_to_open_doors
        ):  # breaks the integration if the water level difference is smaller than a default 5 cm (the last 5 cm of water level difference takes long to overcome, so lock master opens doors)
            z[(i + 1) :] = np.nan  # set all next values of the water level series to nan
            break

    # determining levelling time based on the first nan of the series TODO: Class-functie maken _determine_levelling_time()
    if len(np.argwhere(np.isnan(z))):
        levelling_time = t[np.argwhere(np.isnan(z))[0]][0]
    else:
        levelling_time = t[-1]

    return levelling_time, t, z



import pandas as pd


def get_vessels_during_leveling(lock, vessels: list) -> list:
    """
    Identifies which vessels were present during each lock leveling event.

    Parameters:
    - lock: An object with a `.logbook` attribute (list of dicts with 'Message' and 'Timestamp').
    - vessels: List of vessel objects, each with a `.logbook` attribute and optional `.name`.

    Returns:
    - List of dicts with keys: 'leveling_start', 'leveling_stop', 'vessels_present'
    """
    # Convert lock logbook to DataFrame
    lock_df = pd.DataFrame(lock.logbook)

    # Extract leveling start/stop events
    leveling_starts = lock_df[lock_df["Message"] == "Lock chamber converting start"]
    leveling_stops = lock_df[lock_df["Message"] == "Lock chamber converting stop"]

    # Ensure matching pairs
    leveling_events = pd.DataFrame({
        "leveling_start": leveling_starts["Timestamp"].values,
        "leveling_stop": leveling_stops["Timestamp"].values
    })

    leveling_cycles = []
    for _, level_event in leveling_events.iterrows():
        vessels_present = []
        for vessel in vessels:
            vessel_df = pd.DataFrame(vessel.logbook)
            vessel_df["Timestamp"] = pd.to_datetime(vessel_df["Timestamp"])
            name = getattr(vessel, "name", f"Vessel_{vessels.index(vessel)+1}")

            # Find levelling start/stop pairs
            levelling_starts = vessel_df[vessel_df["Message"] == "Levelling start"]
            levelling_stops = vessel_df[vessel_df["Message"] == "Levelling stop"]

            for i in range(min(len(levelling_starts), len(levelling_stops))):
                start = levelling_starts.iloc[i]["Timestamp"]
                stop = levelling_stops.iloc[i]["Timestamp"]
                if start <= level_event["leveling_stop"] and stop >= level_event["leveling_start"]:
                    vessels_present.append(name)
                    break

        leveling_cycles.append({
            "leveling_start": level_event["leveling_start"],
            "leveling_stop": level_event["leveling_stop"],
            "vessels_present": vessels_present
        })

    return leveling_cycles

def calculate_cycle_looptimes(leveling_cycles: list, vessels: list) -> pd.DataFrame:
    """
    Calculates the looptime for each locking cycle.

    Looptime is defined as the time between:
    - the last vessel exiting the lock in the previous cycle ('Sailing to lock complex exit start')
    - the first vessel entering the lock in the current cycle ('Sailing to first lock doors stop')

    Parameters:
    - leveling_cycles: List of dicts from get_vessels_during_leveling, each with 'vessels_present'.
    - vessels: List of vessel objects, each with a `.logbook` attribute and optional `.name`.

    Returns:
    - DataFrame with columns: 'cycle', 'looptime_seconds'
    """
    # Create a lookup for vessel logbooks
    vessel_logs = {}
    for vessel in vessels:
        name = getattr(vessel, "name", f"Vessel_{vessels.index(vessel)+1}")
        vessel_logs[name] = vessel.logbook

    results = []
    for i, cycle in enumerate(leveling_cycles):
        if i == 0:
            results.append({
                "cycle": i + 1,
                "looptime_seconds": 0
            })
            continue

        prev_vessels = leveling_cycles[i - 1]["vessels_present"]
        curr_vessels = cycle["vessels_present"]

        # Get latest exit time from previous cycle
        prev_exit_times = [
            event["Timestamp"]
            for v in prev_vessels
            for event in vessel_logs.get(v, [])
            if event["Message"] == "Sailing to lock complex exit start"
        ]
        last_exit = max(prev_exit_times) if prev_exit_times else None

        # Get earliest entry time from current cycle
        curr_entry_times = [
            event["Timestamp"]
            for v in curr_vessels
            for event in vessel_logs.get(v, [])
            if event["Message"] == "Sailing to first lock doors stop"
        ]
        first_entry = min(curr_entry_times) if curr_entry_times else None

        # Calculate looptime
        looptime = (first_entry - last_exit).total_seconds() if last_exit and first_entry else None

        results.append({
            "cycle": i + 1,
            "looptime_seconds": looptime
        })

    return pd.DataFrame(results)


def calculate_detailed_cycle_time(lock, vessels, leveling_cycles):
    """
    Calculates detailed timing metrics for full lock cycles (up + down) using logbook data.

    Each full cycle consists of two consecutive leveling events: one upward and one downward.
    The function computes:
    - Looptimes before each phase (t_l_up, t_l_down)
    - Entry and exit durations for vessels based on first and last movement timestamps
    - Lock operation durations (door opening/closing, water level adjustment) per cycle
    - Total cycle time (Tc)
    - Locking system intensity (I_s = 2 * n_max / (Tc / 3600))

    Parameters
    ----------
    lock : object
        Lock object with a `.logbook` attribute containing a list of dicts with 'Message' and 'Timestamp'.
    vessels : list
        List of vessel objects, each with a `.logbook` attribute and optional `.name`.
    leveling_cycles : list
        Output from `get_vessels_during_leveling`, containing dicts with 'leveling_start', 'leveling_stop', and 'vessels_present'.

    Returns
    -------
    pd.DataFrame
        A DataFrame with one row per full lock cycle and the following columns:
        - t_l_up: Looptime before upcycle
        - sum_t_i_up: Time between first vessel's entry start and last vessel's entry stop (upcycle)
        - T_close_up, T_waterlevel_up, T_open_up: Lock operation durations (upcycle)
        - sum_t_u_up: Time between first vessel's exit start and last vessel's exit stop (upcycle)

        - t_l_down: Looptime before downcycle
        - sum_t_i_down: Time between first vessel's entry start and last vessel's entry stop (downcycle)
        - T_close_down, T_waterlevel_down, T_open_down: Lock operation durations (downcycle)
        - sum_t_u_down: Time between first vessel's exit start and last vessel's exit stop (downcycle)

        - Tc_seconds: Total time for the full cycle (up + down)
        - up_vessels: List of vessels in upcycle
        - down_vessels: List of vessels in downcycle
        - I_s: Locking system intensity (vessels per hour)

    Notes
    -----
    - Lock operation durations are extracted per cycle from the lock's logbook using start/stop message pairs.
    - Vessel entry and exit durations are calculated using the earliest and latest timestamps of relevant movement messages.
    - The function assumes `leveling_cycles` are chronologically ordered and alternate between up and down phases.
    """

    def get_duration(df, start_msg, stop_msg):
        starts = df[df["Message"] == start_msg]["Timestamp"].reset_index(drop=True)
        stops = df[df["Message"] == stop_msg]["Timestamp"].reset_index(drop=True)
        return [(stop - start).total_seconds() for start, stop in zip(starts, stops)]

    def get_time_range(log, start_msg, stop_msg):
        df = pd.DataFrame(log)
        df["Timestamp"] = pd.to_datetime(df["Timestamp"])
        starts = df[df["Message"] == start_msg]["Timestamp"]
        stops = df[df["Message"] == stop_msg]["Timestamp"]
        if not starts.empty and not stops.empty:
            return starts.iloc[0], stops.iloc[-1]
        return None, None

    lock_df = pd.DataFrame(lock.logbook)
    lock_df["Timestamp"] = pd.to_datetime(lock_df["Timestamp"])

    # Extract per-cycle lock durations
    T_close_list = get_duration(lock_df, "Lock doors closing start", "Lock doors closing stop")
    T_waterlevel_list = get_duration(lock_df, "Lock chamber converting start", "Lock chamber converting stop")
    T_open_list = get_duration(lock_df, "Lock doors opening start", "Lock doors opening stop")

    vessel_logs = {
        getattr(v, "name", f"Vessel_{i + 1}"): v.logbook
        for i, v in enumerate(vessels)
    }

    results = []
    for i in range(0, len(leveling_cycles) - 1, 2):
        up_cycle = leveling_cycles[i]
        down_cycle = leveling_cycles[i + 1]

        up_vessels = up_cycle["vessels_present"]
        down_vessels = down_cycle["vessels_present"]

        # t_l_up
        if i == 0:
            t_l_up = 0
        else:
            prev_down_vessels = leveling_cycles[i - 1]["vessels_present"]
            last_exit_prev = max([
                get_time_range(vessel_logs[v], "Sailing to second lock doors start",
                               "Sailing to second lock doors stop")[1]
                for v in prev_down_vessels if v in vessel_logs
            ], default=None)
            first_entry_up = min([
                get_time_range(vessel_logs[v], "Sailing to first lock doors stop", "Sailing to first lock doors stop")[
                    0]
                for v in up_vessels if v in vessel_logs
            ], default=None)
            t_l_up = (first_entry_up - last_exit_prev).total_seconds() if first_entry_up and last_exit_prev else 0

        # t_l_down
        last_exit_up = max([
            get_time_range(vessel_logs[v], "Sailing to second lock doors start", "Sailing to second lock doors stop")[1]
            for v in up_vessels if v in vessel_logs
        ], default=None)
        first_entry_down = min([
            get_time_range(vessel_logs[v], "Sailing to first lock doors stop", "Sailing to first lock doors stop")[0]
            for v in down_vessels if v in vessel_logs
        ], default=None)
        t_l_down = (first_entry_down - last_exit_up).total_seconds() if first_entry_down and last_exit_up else 0

        # Entry and exit durations using time range
        entry_start_up, entry_stop_up = None, None
        exit_start_up, exit_stop_up = None, None
        entry_start_down, entry_stop_down = None, None
        exit_start_down, exit_stop_down = None, None

        entry_times_up = [
            get_time_range(vessel_logs[v], "Sailing to position in lock start", "Sailing to position in lock stop") for
            v in up_vessels if v in vessel_logs]
        exit_times_up = [
            get_time_range(vessel_logs[v], "Sailing to second lock doors start", "Sailing to second lock doors stop")
            for v in up_vessels if v in vessel_logs]
        entry_times_down = [
            get_time_range(vessel_logs[v], "Sailing to position in lock start", "Sailing to position in lock stop") for
            v in down_vessels if v in vessel_logs]
        exit_times_down = [
            get_time_range(vessel_logs[v], "Sailing to second lock doors start", "Sailing to second lock doors stop")
            for v in down_vessels if v in vessel_logs]

        # Sum of entering times (up), - Part III, Ch 3, Eq. 3.2
        entry_start_up = min([t[0] for t in entry_times_up if t[0] is not None], default=None)
        entry_stop_up = max([t[1] for t in entry_times_up if t[1] is not None], default=None)
        sum_t_i_up = (entry_stop_up - entry_start_up).total_seconds() if entry_start_up and entry_stop_up else 0

        # Sum of exiting times (up), - Part III, Ch 3, Eq. 3.4
        exit_start_up = min([t[0] for t in exit_times_up if t[0] is not None], default=None)
        exit_stop_up = max([t[1] for t in exit_times_up if t[1] is not None], default=None)
        sum_t_u_up = (exit_stop_up - exit_start_up).total_seconds() if exit_start_up and exit_stop_up else 0

        # Sum of entering times (down), - Part III, Ch 3, Eq. 3.2
        entry_start_down = min([t[0] for t in entry_times_down if t[0] is not None], default=None)
        entry_stop_down = max([t[1] for t in entry_times_down if t[1] is not None], default=None)
        sum_t_i_down = (
                    entry_stop_down - entry_start_down).total_seconds() if entry_start_down and entry_stop_down else 0

        # Sum of exiting times (down), - Part III, Ch 3, Eq. 3.4
        exit_start_down = min([t[0] for t in exit_times_down if t[0] is not None], default=None)
        exit_stop_down = max([t[1] for t in exit_times_down if t[1] is not None], default=None)
        sum_t_u_down = (exit_stop_down - exit_start_down).total_seconds() if exit_start_down and exit_stop_down else 0

        # Identify the index of the op and down cycles
        cycle_index_up = i
        cycle_index_down = i + 1

        # Operation components (up) - Part III, Ch 3, Eq. 3.3
        T_close_up = T_close_list[cycle_index_up] if cycle_index_up < len(T_close_list) else 0
        T_waterlevel_up = T_waterlevel_list[cycle_index_up] if cycle_index_up < len(T_waterlevel_list) else 0
        T_open_up = T_open_list[cycle_index_up] if cycle_index_up < len(T_open_list) else 0

        # Operation components (down) - Part III, Ch 3, Eq. 3.3
        T_close_down = T_close_list[cycle_index_down] if cycle_index_down < len(T_close_list) else 0
        T_waterlevel_down = T_waterlevel_list[cycle_index_down] if cycle_index_down < len(T_waterlevel_list) else 0
        T_open_down = T_open_list[cycle_index_down] if cycle_index_down < len(T_open_list) else 0

        # Part III, Ch 3, Eq. 3.1
        Tc_seconds = (
                t_l_up + sum_t_i_up + T_close_up + T_waterlevel_up + T_open_up + sum_t_u_up +
                t_l_down + sum_t_i_down + T_close_down + T_waterlevel_down + T_open_down + sum_t_u_down
        )

        # Part III, Ch 3, Eq. 3.6
        # NB: n_max is here equivalent to 2 * n_max, since it counts both the up and down vessels
        n_max = len(up_vessels) + len(down_vessels)
        I_s = (n_max / (Tc_seconds / 3600)) if Tc_seconds else None

        results.append({
            "t_l_up": t_l_up,
            "sum_t_i_up": sum_t_i_up,
            "T_close_up": T_close_up,
            "T_waterlevel_up": T_waterlevel_up,
            "T_open_up": T_open_up,
            "sum_t_u_up": sum_t_u_up,
            "t_l_down": t_l_down,
            "sum_t_i_down": sum_t_i_down,
            "T_close_down": T_close_down,
            "T_waterlevel_down": T_waterlevel_down,
            "T_open_down": T_open_down,
            "sum_t_u_down": sum_t_u_down,
            "Tc_seconds": Tc_seconds,
            "up_vessels": up_vessels,
            "down_vessels": down_vessels,
            "I_s": I_s
        })

    return pd.DataFrame(results)


"""Module to load the fairway information system (FIS) network."""

# packkage(s) for documentation, debugging, saving and loading
import io
import pickle

import requests
import requests_cache

# spatial libraries
import shapely.geometry

# inject caching
requests_cache.install_cache("fis_cache")

fis_urls = {
    "0.2": "https://zenodo.org/record/4578289/files/network_digital_twin_v0.2.pickle",
    "0.3": "https://zenodo.org/record/6673604/files/network_digital_twin_v0.3.pickle",
}
euris_urls = {
    "0.1": "https://zenodo.org/records/17298014/files/export-graph-v0.1.0.pickle"
}
networks = {
    "fis": fis_urls,
    "euris": euris_urls,
}

def load_network(network="fis", version="0.3"):
    """load the pickle version of the fairway information system network

    Parameters
    ----------
    network : str
        The network to load. Choose "fis" or "euris". Default is "fis".
    version : str
        The version of the network to load. Choose 0.2 or 0.3. Default is "0.3".
    """
    urls = networks[network]
    url = urls[version]
    resp = requests.get(url)
    # convert the response to a file
    f = io.BytesIO(resp.content)

    # read the graph
    graph = pickle.load(f)

    # convert the edges and nodes geometry to shapely objects
    for e in graph.edges:
        edge = graph.edges[e]
        edge["geometry"] = shapely.geometry.shape(edge["geometry"])
    for n in graph.nodes:
        node = graph.nodes[n]
        node["geometry"] = shapely.geometry.shape(node["geometry"])

    return graph



"""
Utilities for OpenTNSim

This file also includes the networkx read shapefile functions that used to be in networkx.
These originate from:
https://github.com/networkx/networkx/blob/9256ef670730b741105a9264197353004bd6719f/networkx/readwrite/nx_shp.py

Generates a networkx.DiGraph from point and line shapefiles.

"The Esri Shapefile or simply a shapefile is a popular geospatial vector
data format for geographic information systems software. It is developed
and regulated by Esri as a (mostly) open specification for data
interoperability among Esri and other software products."
See https://en.wikipedia.org/wiki/Shapefile for additional information.
"""

# packkage(s) for documentation, debugging, saving and loading
import pathlib
import warnings

# spatial libraries
import networkx as nx

# time libraries
import datetime

# data libraries
import numpy as np
import pandas as pd

# OpenTNSim
import opentnsim


def inherit_docstring(cls):
    for name, func in cls.__dict__.items():
        if callable(func) and not func.__doc__:
            parent_func = getattr(super(cls, cls), name, None)
            if parent_func:
                func.__doc__ = parent_func.__doc__
    return cls


def find_notebook_path():
    """Lookup the path where the notebooks are located. Returns a pathlib.Path object."""
    opentnsim_path = pathlib.Path(opentnsim.__file__)
    # check if the path looks normal
    assert "opentnsim" in str(opentnsim_path), "we can't find the opentnsim path: {opentnsim_path} (opentnsim not in path name)"
    # src_dir/opentnsim/__init__.py -> ../.. -> src_dir
    notebook_path = opentnsim_path.parent.parent / "notebooks"
    return notebook_path


def time_to_numpy(t_start):
    """Convert time to np.datetime64

    Parameters
    ----------
    t_start : float, datetime.datetime, pd.Timestamp
        the time to be converted
    Returns
    -------
    t_start : np.datetime64
        the converted time
    """
    if isinstance(t_start, float):
        t_start = np.datetime64(datetime.datetime.fromtimestamp(t_start))
    elif isinstance(t_start, datetime.datetime):
        t_start = np.datetime64(t_start)
    elif isinstance(t_start, pd.Timestamp):
        t_start = np.array([t_start], dtype=np.datetime64)[0]
    return t_start


# // END-NOSCAN



"""Sailing strategies in OpenTNSim.

The vessel-waterway interaction is complex, since the vessel's sailing behaviors are influenced
by not only the changing waterway situations along the route, but also the varation of its own payload,
fuel weight, velocity, engine power, etc. Therefore, to sail wisely and safely, it is necessary to
formulate optimal sailing strategies to achieve various sailing goals according to different needs.

This package combine with "optimal sailing strategies notebook" provides strategies for
 preventing ship grounding, optimizing cargo capacity, optimizing fuel usage, reducing emissions,
 considering sailing duration, etc.
"""

# To Do in this package:
# 1) add "burning lighter" function to monitor the fuel weight decreasing along the route.
# For the battery-container or electricity powered vessel, the "fuel weight" is constant.
# 2）add "refueling heavier" function to show the fuel weight increased again at the refueling
# stations. For the battery-container or electricy powered vessel, the "fuel weight" is constant.
# 3) add "get_fuel_weight" function which call both the "burning lighter" and "refueling heavier"
# functions to take into account the influence of the variation of fuel weight to the actual draught and payload.
# 4) add "get_refueling_duration" function. For the battery-container, the duration is the
# unloading and loading time for the battery-containers.
# 5) add "get_optimal_refueling_amount" function. It's not always beneficial to be fully refueled
# with fuel for sailing, since more fuel on board leads to less cargo and there might still be
# residual fuel in the tank after a round trip if refuel too much. Therefore, it's needed to
# calculate the optimal refuling amount for each unique sailing case
# (route, vessel size & type, payload, time plan, refueling spots along the route).
# The optimal refueling amount for a transport case determined by both the fuel consumption in
# time and space and the locations of refuling spots.
# 6) consider writting the "fix power or fix speed" example which is in the paper as a function
# into this pacakge in the future or Figure 10 -12 notebooks are enough already? What might it
# benefit if adds this function?

# mathematical packages
import functools
import itertools

# packkage(s) for documentation, debugging, saving and loading
import logging

# package(s) for data handling
import numpy as np
import pandas as pd
import scipy.optimize
import tqdm

logger = logging.getLogger(__name__)


# To know the corresponding Payload for each T_strategy
def T2Payload(vessel, T_strategy, vessel_type):
    """Calculate the corresponding payload for each T_strategy
    the calculation is based on Van Dorsser et al's method (2020)
    (https://www.researchgate.net/publication/344340126_The_effect_of_low_water_on_loading_capacity_of_inland_ships)


    input:
    - T_strategy: user given possible draught
    - vessel types: "Container","Dry_SH","Dry_DH","Barge","Tanker".
    ("Dry_SH" means dry bulk single hull, "Dry_DH" means dry bulk double hull)

    output:
    - Payload_comupted: corresponding payload for the T_strategy for different vessel types

    """
    # Design draft T_design, refer to Table 5

    Tdesign_coefs = dict(
        {
            "intercept": 0,
            "c1": 1.7244153371,
            "c2": 2.2767179246,
            "c3": 1.3365379898,
            "c4": -5.9459308905,
            "c5": 6.2902305560 * 10**-2,
            "c6": 7.7398861528 * 10**-5,
            "c7": 9.0052384439 * 10**-3,
            "c8": 2.8438560877,
        }
    )

    assert vessel_type in [
        "Container",
        "Dry_SH",
        "Dry_DH",
        "Barge",
        "Tanker",
    ], 'Invalid value vessel_type, should be "Container","Dry_SH","Dry_DH","Barge" or "Tanker"'
    if vessel_type == "Container":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [1, 0, 0, 0]
    elif vessel_type == "Dry_SH":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 1, 0, 0]
    elif vessel_type == "Dry_DH":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 1, 0, 0]
    elif vessel_type == "Barge":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 0, 1, 0]
    elif vessel_type == "Tanker":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 0, 0, 1]

    T_design = (
        Tdesign_coefs["intercept"]
        + (dum_container * Tdesign_coefs["c1"])
        + (dum_dry * Tdesign_coefs["c2"])
        + (dum_barge * Tdesign_coefs["c3"])
        + (dum_tanker * Tdesign_coefs["c4"])
        + (Tdesign_coefs["c5"] * dum_container * vessel.L**0.4 * vessel.B**0.6)
        + (Tdesign_coefs["c6"] * dum_dry * vessel.L**0.7 * vessel.B**2.6)
        + (Tdesign_coefs["c7"] * dum_barge * vessel.L**0.3 * vessel.B**1.8)
        + (Tdesign_coefs["c8"] * dum_tanker * vessel.L**0.1 * vessel.B**0.3)
    )

    # Empty draft T_empty, refer to Table 4
    Tempty_coefs = dict(
        {
            "intercept": 7.5740820927 * 10**-2,
            "c1": 1.1615080992 * 10**-1,
            "c2": 1.6865973494 * 10**-2,
            "c3": -2.7490565381 * 10**-2,
            "c4": -5.1501240744 * 10**-5,
            "c5": 1.0257551153 * 10**-1,
            "c6": 2.4299435211 * 10**-1,
            "c7": -2.1354295627 * 10**-1,
        }
    )

    if vessel_type == "Container":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [1, 0, 0, 0]
    elif vessel_type == "Dry_SH":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 0, 0, 0]
    elif vessel_type == "Dry_DH":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 1, 0, 0]
    elif vessel_type == "Barge":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 0, 1, 0]

    elif vessel_type == "Tanker":
        [dum_container, dum_dry, dum_barge, dum_tanker] = [0, 0, 0, 1]

    # dum_container and dum_dry use the same "c5"
    T_empty = (
        Tempty_coefs["intercept"]
        + (Tempty_coefs["c1"] * vessel.B)
        + (Tempty_coefs["c2"] * ((vessel.L * T_design) / vessel.B))
        + (Tempty_coefs["c3"] * (np.sqrt(vessel.L * vessel.B)))
        + (Tempty_coefs["c4"] * (vessel.L * vessel.B * T_design))
        + (Tempty_coefs["c5"] * dum_container)
        + (Tempty_coefs["c5"] * dum_dry)
        + (Tempty_coefs["c6"] * dum_tanker)
        + (Tempty_coefs["c7"] * dum_barge)
    )

    # Capacity indexes, refer to Table 3 and eq 2
    CI_coefs = dict(
        {
            "intercept": 2.0323139721 * 10**1,
            "c1": -7.8577991460 * 10**1,
            "c2": -7.0671612519 * 10**0,
            "c3": 2.7744056480 * 10**1,
            "c4": 7.5588609922 * 10**-1,
            "c5": 3.6591813315 * 10**1,
        }
    )
    # Capindex_1 related to actual draft (especially used for shallow water)
    Capindex_1 = (
        CI_coefs["intercept"]
        + (CI_coefs["c1"] * T_empty)
        + (CI_coefs["c2"] * T_empty**2)
        + (CI_coefs["c3"] * T_strategy)
        + (CI_coefs["c4"] * T_strategy**2)
        + (CI_coefs["c5"] * (T_empty * T_strategy))
    )
    # Capindex_2 related to design draft
    Capindex_2 = (
        CI_coefs["intercept"]
        + (CI_coefs["c1"] * T_empty)
        + (CI_coefs["c2"] * T_empty**2)
        + (CI_coefs["c3"] * T_design)
        + (CI_coefs["c4"] * T_design**2)
        + (CI_coefs["c5"] * (T_empty * T_design))
    )

    # DWT design capacity, refer to Table 6 and eq 3
    capacity_coefs = dict(
        {
            "intercept": -1.6687441313 * 10**1,
            "c1": 9.7404521380 * 10**-1,
            "c2": -1.1068568208,
        }
    )

    DWT_design = (
        capacity_coefs["intercept"]
        + (capacity_coefs["c1"] * vessel.L * vessel.B * T_design)
        + (capacity_coefs["c2"] * vessel.L * vessel.B * T_empty)
    )  # designed DWT
    DWT_actual = (Capindex_1 / Capindex_2) * DWT_design  # actual DWT of shallow water

    if T_strategy < T_design:
        DWT_final = DWT_actual
        # Consumables represents the persentage of fuel weight,which is 4-6% of designed DWT
        # 4% for shallow water (Van Dosser  et al. Chapter 8,pp.68).
        # Based on personal communication with experts we lowered this to 0.005.
        # This should match with the current practice.
        consumables = 0.005

    else:
        DWT_final = DWT_design
        # consumables represents the persentage of fuel weight,which is 4-6% of designed DWT
        # 4% for shallow water (Van Dosser  et al. Chapter 8,pp.68).
        # Based on personal communication with experts we lowered this to 0.005.
        # This should match with the current practice.
        consumables = 0.005

    fuel_weight = DWT_design * consumables  # (Van Dosser et al. Chapter 8, pp.68).

    Payload_computed = DWT_final - fuel_weight  # payload=DWT-fuel_weight

    # DWT_final covers the situations of the DWT at maximum draught and the DWT at adjusted draught
    # We include DWT_final for calculating cargo-fuel trade off by function 'get_adjusted_cargo_amount'.
    return Payload_computed


def Payload2T(vessel, Payload_strategy, vessel_type, bounds=(0, 5)):
    """Calculate the corresponding draught (T_Payload2T) for each Payload_strategy
    the calculation is based on Van Dorsser et al's method (2020)
    (https://www.researchgate.net/publication/344340126_The_effect_of_low_water_on_loading_capacity_of_inland_ships),
      which applyies for inland shipping.


    input:
    - Payload_strategy: user given payload
    - vessel types: "Container","Dry_SH","Dry_DH","Barge","Tanker".
    ("Dry_SH" means dry bulk single hull, "Dry_DH" means dry bulk double hull)
    - bounds: the searching range for draught. As this method which based on
    Van Dorsser et al (2020) is for inland vessels, of which the draughts are no
    larger than 5 meter, we set the upper bound as 5 m as default value.

    output:
    - T_Payload2T: corresponding draught for each payload for different vessel types

    """

    def seek_T_given_Payload(T_Payload2T, vessel, vessel_type):
        """function to optimize"""
        Payload_computed = T2Payload(vessel=vessel, T_strategy=T_Payload2T, vessel_type=vessel_type)
        # compute difference between a given payload (Payload_strategy) and a computed payload (Payload_computed)
        diff = Payload_strategy - Payload_computed
        return diff**2

    # fill in some of the parameters that we already know
    fun = functools.partial(seek_T_given_Payload, vessel=vessel, vessel_type=vessel_type)

    # lookup a minimum
    fit = scipy.optimize.minimize_scalar(fun, bounds=bounds, method="bounded")

    # check if we found a minimum
    if not fit.success:
        raise ValueError(fit)

    # the value of fit.x within the bound (0,5) is the draught we find where the diff**2 reach a minimum (zero).
    T_Payload2T = fit.x

    return T_Payload2T


def get_v(vessel, width, depth, margin, bounds):
    """for a waterway section with a given width and depth, compute the velocity that can be
    reached given a vessel's T and a safety margin."""

    def seek_v_given_z(v, vessel, width, depth, margin):

        # calculate the sinkage (if turned on, otherwise 0)
        max_sinkage = vessel.calculate_max_sinkage(v=v, h_0=depth, width=width)

        # calculate available underkeel clearance (vessel in rest)
        available_clearance = depth - vessel._T

        # compute difference between the sinkage and the space available for sinkage (including safety margin)
        diff = available_clearance - max_sinkage - margin

        return diff**2

    # goalseek to minimize
    fun = functools.partial(seek_v_given_z, vessel=vessel, width=width, depth=depth, margin=margin)
    fit = scipy.optimize.minimize_scalar(fun, bounds=bounds, method="bounded")

    # check if we found a minimum
    if not fit.success:
        raise ValueError(fit)

    # the value of fit.x within the bound (0,20) is the velocity we find where the diff**2 reach a minimum (zero).
    v = fit.x

    return v, depth, margin



def _power_difference(v, vessel, width, depth):
    """
    Helper function for the SciPy solver.
    It calculates the difference: P_installed - P_tot(v).
    The solver will find the root (where this function is zero).
    
    Note: This function modifies the vessel object's state (P_tot, etc.)
    as a side effect of the calculation methods.
    """
    # Handle v=0 case or very small v
    if v <= 1e-6:
        # At zero velocity, required power is zero.
        # Difference is just P_installed (assuming P_installed > 0)
        return vessel.P_installed - 0.0

    # 1. Calculate squatted water depth for the given velocity
    # in energy module waterdepth is called h_0
    h_0 = vessel.calculate_h_squat(v=v, h_0=depth, width=width)

    # 2. For the squatted water depth, calculate resistance and power
    vessel.calculate_total_resistance(v=v, h_0=h_0)
    vessel.calculate_total_power_required(v=v, h_0=h_0)

    # 3. Return the difference
    return vessel.P_installed - vessel.P_tot


def get_upperbound_for_power2v_optim(vessel, width, depth, margin=0, bounds=(0, 20)):
    """
    For a waterway section with a given width and depth, compute a maximum installed-
    power-allowed velocity, considering squat. (Optimized with SciPy solver)

    This velocity can be used as an upperbound in the power2v function.
    It finds the velocity 'v' where P_tot(v) == P_installed.
    """

    # 1. Estimate the grounding velocity (hard upper limit)
    # This is the absolute maximum velocity before the vessel grounds.
    grounding_v, _, _ = get_v(vessel, width, depth, margin=0, bounds=bounds)

    # If grounding velocity is effectively zero, vessel can't move
    if grounding_v <= 1e-6:
        return 0.0

    # Define a small positive velocity as the lower bound for the solver
    v_min = 1e-6 
    
    # 2. Check the power difference at the lower bound (v_min)
    # We expect P_installed > P_tot, so diff_at_min should be positive
    diff_at_min = _power_difference(v_min, vessel, width, depth)

    if diff_at_min < 0:
        # This means required power is *already* > installed power at min velocity.
        # The vessel cannot move at this speed.
        print(f"Warning: Required power exceeds installed power at min velocity ({v_min} m/s).")
        return 0.0

    # 3. Check the power difference at the upper bound (grounding_v)
    diff_at_grounding = _power_difference(grounding_v, vessel, width, depth)

    if diff_at_grounding >= 0:
        # P_installed >= P_tot even at the grounding velocity.
        # This means the vessel is limited by *grounding*, not by power.
        # The maximum allowed velocity is the grounding velocity.
        return grounding_v

    # 4. We have a valid bracket for the solver:
    # diff_at_min > 0 (power is available)
    # diff_at_grounding < 0 (power limit exceeded)
    # The solver can now find the root v where diff == 0.
    
    sol = scipy.optimize.root_scalar(
        _power_difference,       # The function to solve
        args=(vessel, width, depth), # Extra args to pass to the function
        method='brentq',         # Use Brent's method (fast and robust)
        bracket=[v_min, grounding_v] # Search between min and grounding velocity
    )

    if sol.converged:
        # The solver found the velocity where P_installed == P_tot
        return sol.root
    else:
        # Solver failed, which is rare for brentq with a valid bracket.
        # Fallback to a safe value.
        print(f"Warning: Power-velocity solver did not converge. Returning 0.")
        return 0.0

def get_upperbound_for_power2v(vessel, width, depth, margin=0, bounds=(0, 20)):
    """for a waterway section with a given width and depth, compute a maximum installed-
    power-allowed velocity, considering squat. This velocity can be used as upperbound in the
    power2v function in energy.py. The "upperbound" is the maximum value in velocity searching
    range."""

    # estimate the grounding velocity
    # here we optionally try to take sinkage into account and try to compute the maximum velocity
    # where we still have underkeel clearance
    grounding_v, depth, margin = get_v(vessel, width, depth, margin=0, bounds=bounds)

    # The next step is to compute the maximum power for the veloctiy.
    # Here (for some reason we don't use a solver )
    # find max power velocity
    velocity = np.linspace(0.01, grounding_v, 500)

    # Here we create a list of velocities from 0.01 to the grounding velocity in 100 steps
    # For each velocity we will calculate the squat and the waterdepth after squat
    task = list(itertools.product(velocity[0:-1]))

    # prepare a list of dictionaries for pandas
    rows = []
    for item in task:
        row = {"velocity": item[0]}
        rows.append(row)

    # convert simulations to dataframe, so that we can apply a function and monitor progress
    task_df = pd.DataFrame(rows)

    # creat a results empty list to collect the below results
    # TODO: replace this with a goalseeking function, now we need 500 iterations to get an accurate result
    # With a goal seeking function (similar to get_v, we should have this down to ~10-20)
    # Or we can compute it analytically, but that requres a lot of coding
    results = []
    for i, row in tqdm.tqdm(task_df.iterrows(), disable=True):

        # in energy module waterdepth  is called h_0 (ignores when squat is disabled)
        h_0 = vessel.calculate_h_squat(v=row["velocity"], h_0=depth, width=width)
        max_sinkage = vessel.calculate_max_sinkage(v=row["velocity"], h_0=depth, width=width)

        # for the squatted water depth calculate resistance and power
        vessel.calculate_total_resistance(v=row["velocity"], h_0=h_0)
        vessel.calculate_total_power_required(v=row["velocity"], h_0=h_0)

        # prepare a row
        result = {}
        result.update(row)
        result["Powerallowed_v"] = row["velocity"]
        result["P_tot"] = vessel.P_tot
        result["P_installed"] = vessel.P_installed
        # not used but for inspection
        result["depth"] = depth
        result["max_sinkage"] = max_sinkage
        result["h_0"] = h_0

        # update resulst dict
        results.append(result)

    results_df = pd.DataFrame(results)

    # return only dataframe up to and including the first time that P_tot == P_installed
    # TODO: is P_tot again incl hotel power, do we want to focus on P_tot being equal to P_available_for_propulsion
    selected = results_df[0 : (results_df.P_installed >= results_df.P_tot).idxmin()]

    upperbound = selected.Powerallowed_v.max()

    return upperbound, selected, results_df



# -*- coding: utf-8 -*-

"""Console script for opentnsim."""
# package(s) related to the command line interface
import sys

import click

import opentnsim.server


@click.group()
def cli(args=None):
    """OpenTNSim simulation."""
    click.echo("Replace this message by putting your code into " "opentnsim.cli.main")
    click.echo("See click documentation at http://click.pocoo.org/")
    return 0


@cli.command()
@click.option("--host", default="0.0.0.0")
@click.option("--port", default=5000, type=int)
@click.option("--debug/--no-debug", default=False)
def serve(host, port, debug, args=None):
    """Run a flask server with the backend code"""
    app = opentnsim.server.app
    app.run(host=host, debug=debug, port=port)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



"""Module with calculation functions for the energy module
Most functions are called by the ConsumesEnergy mixin.

"""

# %% IMPORT DEPENDENCIES
# generic
#
import collections

import logging
import numpy as np

# logging
logger = logging.getLogger(__name__)


def sample_engine_age(L_w):
    """
    Samples the age of the engine based on the weight class of the vessel. The age is
    drawn randomly from a Weibull distribution with parameters k and lmb, which depend
    on the weight class of the ship. The age is then returned in years.

    Ligterink, N.E., R.N. van Gijlswijk, G. Kadijk, R.J. Vermeulen, A.P. Indrajuana, M. Elstgeest,
    P. van Mensch, J.M. de Ruiter, R.P. Verbeek, J.H.J. Hulskotte, G. Geilenkirchen (PBL) and
    M. Traa (PBL) (2019). "Emissiefactoren wegverkeer - Actualisatie 2019", TNO 2019 R10825v2.
    https://www.pbl.nl/publicaties/emissiefactoren-wegverkeer-actualisatie-2019

    Parameters
    ----------
    L_w : int
        weight class of the ship (depending on carrying capacity), values supported are
        1 (class L1), 2 (class L2) or 3 (class L3).

    Returns
    -------
    age : int
        The calculated age of the engine in years.
    """
    # check params
    if not L_w in [1, 2, 3]:
        raise ValueError("L_w should be 1, 2 or 3")

    # determine the shape (k) and scale factor (lmb) to use based on the weight class
    # Table 8, in Ligterink et al (2019)
    if L_w == 1:  # Weight class L1
        k = 1.3
        lmb = 20.4
    elif L_w == 2:  # Weight class L2
        k = 1.12
        lmb = 18.5
    elif L_w == 3:  # Weight class L3
        k = 1.26
        lmb = 18.6

    # the engine age
    # returns a randomly sampled estimate of the engine age (in years) for a ship, based on its weight class
    age = int(np.random.weibull(k) * lmb)

    return age


def calculate_max_sinkage(v, h_0, T, B, C_B, width):
    """
    Calculate tha maximum sinkage of a moving vessel.

    The calculation equation is described in Barrass, B. & Derrett, R.'s book (2006),
    Ship Stability for Masters and Mates, chapter 42.
    https://doi.org/10.1016/B978-0-08-097093-6.00042-6

    Parameters
    ----------
    v : float
        Velocity of the vessel relative to the water [m/s]
    h_0 : float
        Water depth [m]
    T : float
        Actual draught of the vessel [m]
    B : float
        Beam of the vessel [m]
    C_B : float
        Block coefficient of the vessel [-]
    width : float
        Width of the fairway [m]

    Returns
    -------
    float
        The maximum sinkage of the vessel [m]
    """
    # checks
    if v < 0:
        raise ValueError("Velocity v should be >= 0")
    if h_0 <= 0:
        raise ValueError("Water depth h_0 should be > 0")
    if T <= 0:
        raise ValueError("Draught T should be > 0")
    if B <= 0:
        raise ValueError("Beam B should be > 0")
    if width <= 0:
        raise ValueError("Width of the fairway should be > 0")
    if C_B <= 0:
        raise ValueError("Block coefficient C_B should be > 0")

    if B > width:
        raise ValueError(f"Width of the fairway ({width}) should be larger than " f"the beam of the vessel ({B})")

    # calculate the maximum sinkage
    sinkage = (C_B * ((B * T) / (width * h_0)) ** 0.81) * ((v * 1.94) ** 2.08) / 20
    return sinkage


def calculate_properties(C_B, L, B, T, bulbous_bow, C_BB):
    """
    Calculate the properties of a vessel based on its block coefficient, length,
    beam, draught, bulbous bow coefficient, and whether it has a bulbous bow.

    Parameters
    ----------
    C_B : float
        Block coefficient of the vessel [-]
    L : float
        Length of the vessel [m]
    B : float
        Beam of the vessel [m]
    T : float
        Actual draught of the vessel [m]
    bulbous_bow : bool
        Whether the vessel has a bulbous bow or not
    C_BB : float
        Bulbous bow coefficient of the vessel [-]

    Returns
    -------
    C_M : float
        Midship section coefficient [-]
    C_WP : float
        Waterplane coefficient [-]
    C_P : float
        Prismatic coefficient [-]
    delta : float
        Water displacement of the vessel [m^3]
    lcb : float
        Longitudinal center of buoyancy [m]
    L_R : float
        Length parameter reflecting the length of the run [m]
    A_T : float
        Transverse area of the transom [m^2]
    A_BT : float
        Cross-sectional area of the bulb at still water level [m^2]
    S : float
        Total wetted surface area of the vessel [m^2]
    S_APP : float
        Wet area of appendages [m^2]
    S_B : float
        Area of flat bottom [m^2]
    T_F : float
        Forward draught of the vessel [m]
    h_B : float
        Position of the centre of the transverse area [m]
    D_s : float, optional
        Diameter of the screw [m], if not provided, it is assumed to be 0.7 * T

    """

    # TODO: add properties for seagoing ships with bulbs

    # (Van Koningsveld et al (2023) - Part IV Eq 5.9, 5.10 and below Eq 5.12)
    C_M = 1.006 - 0.0056 * C_B ** (-3.56)  # Midship section coefficient (Eq 5.9)
    C_WP = (1 + 2 * C_B) / 3  # Waterplane coefficient (Eq 5.10)
    C_P = C_B / C_M  # Prismatic coefficient (see below Eq 5.12)

    # Segers (2021) (http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3)
    # Appendix C - Eq C.2
    delta = C_B * L * B * T  # Water displacement

    # Van Koningsveld et al (2023) - Part IV Table 5.1
    lcb = -13.5 + 19.4 * C_P  # longitudinal center of buoyancy
    # Van Koningsveld et al (2023) - Part IV Eq 5.13
    L_R = L * (
        1 - C_P + ((0.06 * C_P * lcb) / (4 * C_P - 1)) * (19.4 * C_P - 13.5)
    )  # length parameter reflecting the length of the run

    # Van Koningsveld et al (2023) - below Eq 5.16
    A_T = 0.1 * B * T  # transverse area of the transom
    # calculation for A_BT (cross-sectional area of the bulb at still water level [m^2]) depends on whether a ship has a bulb
    if bulbous_bow:
        # TODO: check Holtrop and Mennen for this formulation
        A_BT = C_BB * B * T * C_M  # calculate A_BT for seagoing ships having a bulb
    else:
        A_BT = 0  # most inland ships do not have a bulb. So we assume A_BT=0.

    # Total wet area: S (Van Koningsveld et al (2023) - Eq 5.8)
    assert C_M >= 0, f"C_M should be positive: {C_M}"
    S = L * (2 * T + B) * np.sqrt(C_M) * (0.453 + 0.4425 * C_B - 0.2862 * C_M - 0.003467 * (B / T) + 0.3696 * C_WP) + 2.38 * (
        A_BT / C_B
    )

    # Segers (2021) (http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3)
    # In the explanation under Eq 3.27
    S_APP = 0.05 * S  # Wet area of appendages
    # Segers (2021) Eq 3.20
    S_B = L * B  # Area of flat bottom

    # TODO: D_s is a property that should be given, not calculated
    # if D_s is None:
    #     D_s = 0.7 * T  # Diameter of the screw

    # TODO: check references for these equations
    T_F = T  # Forward draught of the vessel [m]
    h_B = 0.2 * T  # Position of the centre of the transverse area [m]

    names = ["C_M", "C_WP", "C_P", "delta", "lcb", "L_R", "A_T", "A_BT", "S",
             "S_APP", "S_B", "T_F", "h_B"]
    vessel_energy_properties = collections.namedtuple("VesselEnergyProperties", names)
    result = vessel_energy_properties(C_M=C_M, C_WP=C_WP, C_P=C_P, delta=delta, lcb=lcb, L_R=L_R, A_T=A_T, A_BT=A_BT, S=S, S_APP=S_APP, S_B=S_B, T_F=T_F, h_B=h_B)
    return result


def calculate_frictional_resistance(v, h_0, L, nu, T, S, S_B, rho):
    """
    Calculate the frictional resistance of a vessel.

    This is the first resistance component defined by Holtrop and Mennen (1982). A
    modification to the original friction line is applied, based on literature of
    Zeng (2018), to account for shallow water effects.

    Parameters
    ----------
    v : float
        Velocity of the vessel relative to the water [m/s]
    h_0 : float
        Water depth [m]
    L : float
        Length of the vessel [m]
    nu : float
        Kinematic viscosity of the water [m^2/s]
    T : float
        Actual draught of the vessel [m]
    S : float
        Total wetted surface area of the vessel [m^2]
    S_B : float
        Area of the flat bottom of the vessel [m^2]
    rho : float
        Density of the water [kg/m^3]

    Returns
    -------
    R_f : float
        Frictional resistance of the vessel [kN]
    C_f : float
        Friction coefficient of the vessel [-]
    R_e : float
        Reynolds number of the vessel [-]
    Cf_deep : float
        Friction coefficient in deep water based on CFD computations of
        Zeng et al. (2018) [-]
    Cf_shallow : float
        Friction coefficient in shallow water based on CFD computations of
        Zeng et al. (2018) [-]
    Cf_0 : float
        Friction coefficient in deep water according to ITTC-1957 curve [-]
    Cf_Katsui : float
        Friction coefficient according to Katsui (1978) [-]
    V_B : float
        Average velocity underneath the ship, taking into account the shallow water
        effect [m/s]
    D : float
        Distance from the bottom of the ship to the bottom of the fairway [m]
    a : float
        Coefficient needed to calculate the Katsui friction coefficient [-]
    """
    # Reynolds number
    R_e = v * L / nu

    # distance from bottom ship to the bottom of the fairway
    D = h_0 - T
    if not D > 0:
        raise ValueError(f"Distance between ship and bottom should be > 0: {D}")

    # Friction coefficient based on CFD computations of Zeng et al. (2018), in deep
    # water --> Van Koningsveld et al (2023) - Eq 5.3
    Cf_deep = 0.08169 / ((np.log10(R_e) - 1.717) ** 2)
    assert not isinstance(Cf_deep, complex), f"Cf_deep should not be complex: {Cf_deep}"

    # Friction coefficient based on CFD computations of Zeng et al. (2018), taking into
    # account shallow water effects --> Van Koningsveld et al (2023) - Eq 5.4
    Cf_shallow = (0.08169 / ((np.log10(R_e) - 1.717) ** 2)) * (1 + (0.003998 / (np.log10(R_e) - 4.393)) * (D / L) ** (-1.083))
    assert not isinstance(Cf_shallow, complex), f"Cf_shallow should not be complex: {Cf_shallow}"

    # Friction coefficient in deep water according to ITTC-1957 curve
    # Van Koningsveld et al (2023) - Eq 5.6
    Cf_0 = 0.075 / ((np.log10(R_e) - 2) ** 2)

    # 'a' is the coefficient needed to calculate the Katsui friction coefficient
    # Van Koningsveld et al (2023) - below Eq 5.7
    a = 0.042612 * np.log10(R_e) + 0.56725

    # Van Koningsveld et al (2023) - Eq 5.7
    Cf_Katsui = 0.0066577 / (np.log10(R_e) - 4.3762) ** a

    # The average velocity underneath the ship, taking into account the shallow water
    # effect. This calculation is to get V_B, which will be used in the following Cf
    # for shallow water equation:
    if h_0 / T <= 4:
        V_B = 0.4277 * v * np.exp((h_0 / T) ** (-0.07625))
    else:
        V_B = v

    # cf_shallow and cf_deep cannot be applied directly, since a vessel also has
    # non-horizontal wet surfaces that have to be taken into account. Therefore, the
    # following formula for the final friction coefficient 'C_f' for deep water or
    # shallow water is defined according to Zeng et al. (2018)
    if (h_0 - T) / L > 1:
        # calculate Friction coefficient C_f for deep water:
        # Zeng et al. (2018)
        C_f = Cf_0 + (Cf_deep - Cf_Katsui) * (S_B / S)
        logger.debug("now i am in the deep loop")
    else:
        # calculate Friction coefficient C_f for shallow water:
        # Van Koningsveld et al (2023) - Eq 5.5
        C_f = Cf_0 + (Cf_shallow - Cf_Katsui) * (S_B / S) * (V_B / v) ** 2
        logger.debug("now i am in the shallow loop")
    assert not isinstance(C_f, complex), f"C_f should not be complex: {C_f}"

    # The total frictional resistance R_f [kN]:
    # Van Koningsveld et al (2023) - Eq 5.2
    R_f = (0.5 * rho * (v**2) * C_f * S) / 1000
    assert not isinstance(R_f, complex), f"R_f should not be complex: {R_f}"

    names = ["R_f", "C_f", "R_e", "Cf_deep", "Cf_shallow", "Cf_0", "Cf_Katsui", "V_B", "D", "a"]
    frictional_resisistance_properties = collections.namedtuple("FrictionalResistanceProperties", names)

    result = frictional_resisistance_properties(
        R_f, C_f, R_e, Cf_deep, Cf_shallow, Cf_0, Cf_Katsui, V_B, D, a
    )
    return result


def calculate_viscous_resistance(c_stern, B, L, T, L_R, C_P, R_f, delta):
    """
    Calculate the viscous resistance of a vessel.

    This is the second resistance component defined by Holtrop and Mennen (1982). Form
    factor (1 + k1) has to be multiplied by the frictional resistance R_f, to account
    for the effect of viscosity.

    Parameters
    ----------
    c_stern : float
        Determines the shape of the afterbody [-]
    B : float
        Beam of the vessel [m]
    L : float
        Length of the vessel [m]
    T : float
        Actual draught of the vessel [m]
    L_R : float
        Length parameter reflecting the length of the run [m]
    C_P : float
        Prismatic coefficient of the vessel [-]
    R_f : float
        Frictional resistance of the vessel [kN]
    delta : float
        Water displacement of the vessel [m^3]

    Returns
    -------
    c_14 : float
        Coefficient accounting for the specific shape of the afterbody [-]
    one_k1 : float
        Form factor (1 + k1) describing the viscous resistance [-]
    R_f_one_k1 : float
        Viscous resistance of the vessel, which is the product of the frictional
        resistance R_f and the form factor (1 + k1) [kN]
    """
    # c_14 accounts for the specific shape of the afterbody
    # TODO: check where this value comes from (Holtrop and Mennen?) (following
    #  Segers (2021) we assume c_stern = 0 which leads to c_14 to be 1)
    c_14 = 1 + 0.0011 * c_stern

    # the form factor (1+k1) describes the viscous resistance
    # Van Koningsveld et al (2023) - Eq 5.12
    one_k1 = 0.93 + 0.487 * c_14 * ((B / L) ** 1.068) * ((T / L) ** 0.461) * ((L / L_R) ** 0.122) * (
        ((L**3) / delta) ** 0.365
    ) * ((1 - C_P) ** (-0.604))

    R_f_one_k1 = R_f * one_k1
    viscous_resistance_properties = collections.namedtuple("ViscousResistanceProperties", ["c_14", "one_k1", "R_f_one_k1"])
    result = viscous_resistance_properties(c_14=c_14, one_k1=one_k1, R_f_one_k1=R_f_one_k1)
    return result


def calculate_appendage_resistance(v, rho, S_APP, one_k2, C_f):
    """
    Calculate the frictional resistance resulting from the wetted area of appendages.
    This function computes the appendage resistance (R_APP) in kilonewtons (kN) based
    on the provided parameters, using the formula from Segers (2021) - Eq 3.27.

    Parameters
    ----------
    v : float
        Ship velocity in meters per second (m/s).
    rho : float
        Water density in kilograms per cubic meter (kg/m^3).
    S_APP : float
        Wetted surface area of appendages in square meters (m^2).
    one_k2 : float
        Form factor for appendages (dimensionless).
    C_f : float
        Frictional resistance coefficient (dimensionless).

    Returns
    -------
    float
        Frictional resistance of appendages (R_APP) in kilonewtons (kN).

    References
    ----------
    Segers (2021) - Eq 3.27. http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3
    """
    # Frictional resistance resulting from wetted area of appendages: R_APP [kN]
    # Segers (2021) - Eq 3.27 (http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3)
    R_APP = (0.5 * rho * (v**2) * S_APP * one_k2 * C_f) / 1000

    return R_APP


def karpov(v, h_0, g, T):
    """
    Calculate the corrected velocity and alpha coefficient using the Karpov method.
    The Karpov method applies a correction factor (alpha_xx) to the velocity based on
    the Froude number and the ratio of water depth to draft. The correction is
    determined using polynomial approximations for different ranges of the Froude
    number and h_0/T.

    Parameters
    ----------
    v : float
        The measured velocity (m/s).
    h_0 : float
        The water depth (m).
    g : float
        The gravitational acceleration (m/s^2). Must be non-negative.
    T : float
        The draft (m).

    Returns
    -------
    F_rh : float
        The Froude number based on depth (dimensionless).
    V_2 : float
        The corrected velocity (m/s) according to the Karpov method.
    alpha_xx : float
        The correction coefficient applied to the velocity.


    Notes
    -----
    - The function uses a piecewise polynomial fit for the correction coefficient based
      on the Froude number and h_0/T.
    - The Froude number is calculated as v / sqrt(g * h_0).
    """

    # The Froude number used in the Karpov method is the depth related Froude number F_rh

    # The different alpha** curves are determined with a sixth power polynomial approximation in Excel (Segers, 2019)
    # A distinction is made between different ranges of Froude numbers, because this resulted in a better approximation of the curve

    assert g >= 0, f"g should be positive: {g}"
    assert h_0 >= 0, f"h_0 should be positive: {h_0}"

    F_rh = v / np.sqrt(g * h_0)

    if F_rh <= 0.4:
        if 0 <= h_0 / T < 1.75:
            alpha_xx = (-4 * 10 ** (-12)) * F_rh**3 - 0.2143 * F_rh**2 - 0.0643 * F_rh + 0.9997
        if 1.75 <= h_0 / T < 2.25:
            alpha_xx = -0.8333 * F_rh**3 + 0.25 * F_rh**2 - 0.0167 * F_rh + 1
        if 2.25 <= h_0 / T < 2.75:
            alpha_xx = -1.25 * F_rh**4 + 0.5833 * F_rh**3 - 0.0375 * F_rh**2 - 0.0108 * F_rh + 1
        if h_0 / T >= 2.75:
            alpha_xx = 1
    elif F_rh > 0.4:
        if 0 <= h_0 / T < 1.75:
            alpha_xx = (
                -0.9274 * F_rh**6 + 9.5953 * F_rh**5 - 37.197 * F_rh**4 + 69.666 * F_rh**3 - 65.391 * F_rh**2 + 28.025 * F_rh - 3.4143
            )
        if 1.75 <= h_0 / T < 2.25:
            alpha_xx = (
                2.2152 * F_rh**6 - 11.852 * F_rh**5 + 21.499 * F_rh**4 - 12.174 * F_rh**3 - 4.7873 * F_rh**2 + 5.8662 * F_rh - 0.2652
            )
        if 2.25 <= h_0 / T < 2.75:
            alpha_xx = (
                1.2205 * F_rh**6 - 5.4999 * F_rh**5 + 5.7966 * F_rh**4 + 6.6491 * F_rh**3 - 16.123 * F_rh**2 + 9.2016 * F_rh - 0.6342
            )
        if 2.75 <= h_0 / T < 3.25:
            alpha_xx = (
                -0.4085 * F_rh**6 + 4.534 * F_rh**5 - 18.443 * F_rh**4 + 35.744 * F_rh**3 - 34.381 * F_rh**2 + 15.042 * F_rh - 1.3807
            )
        if 3.25 <= h_0 / T < 3.75:
            alpha_xx = (
                0.4078 * F_rh**6 - 0.919 * F_rh**5 - 3.8292 * F_rh**4 + 15.738 * F_rh**3 - 19.766 * F_rh**2 + 9.7466 * F_rh - 0.6409
            )
        if 3.75 <= h_0 / T < 4.5:
            alpha_xx = (
                0.3067 * F_rh**6 - 0.3404 * F_rh**5 - 5.0511 * F_rh**4 + 16.892 * F_rh**3 - 20.265 * F_rh**2 + 9.9002 * F_rh - 0.6712
            )
        if 4.5 <= h_0 / T < 5.5:
            alpha_xx = (
                0.3212 * F_rh**6 - 0.3559 * F_rh**5 - 5.1056 * F_rh**4 + 16.926 * F_rh**3 - 20.253 * F_rh**2 + 10.013 * F_rh - 0.7196
            )
        if 5.5 <= h_0 / T < 6.5:
            alpha_xx = (
                0.9252 * F_rh**6 - 4.2574 * F_rh**5 + 5.0363 * F_rh**4 + 3.3282 * F_rh**3 - 10.367 * F_rh**2 + 6.3993 * F_rh - 0.2074
            )
        if 6.5 <= h_0 / T < 7.5:
            alpha_xx = (
                0.8442 * F_rh**6 - 4.0261 * F_rh**5 + 5.313 * F_rh**4 + 1.6442 * F_rh**3 - 8.1848 * F_rh**2 + 5.3209 * F_rh - 0.0267
            )
        if 7.5 <= h_0 / T < 8.5:
            alpha_xx = (
                0.1211 * F_rh**6 + 0.628 * F_rh**5 - 6.5106 * F_rh**4 + 16.7 * F_rh**3 - 18.267 * F_rh**2 + 8.7077 * F_rh - 0.4745
            )
        if 8.5 <= h_0 / T < 9.5:
            if F_rh <= 0.6:
                alpha_xx = 1
            if F_rh > 0.6:
                alpha_xx = (
                    -6.4069 * F_rh**6 + 47.308 * F_rh**5 - 141.93 * F_rh**4 + 220.23 * F_rh**3 - 185.05 * F_rh**2 + 79.25 * F_rh - 12.484
                )
        if h_0 / T >= 9.5:
            if F_rh <= 0.6:
                alpha_xx = 1
            if F_rh > 0.6:
                alpha_xx = (
                    -6.0737 * F_rh**6 + 44.97 * F_rh**5 - 135.21 * F_rh**4 + 210.13 * F_rh**3 - 176.72 * F_rh**2 + 75.728 * F_rh - 11.893
                )

    V_2 = v / alpha_xx

    karpov_properties = collections.namedtuple("KarpovProperties", ["F_rh", "V_2", "alpha_xx"])
    result = karpov_properties(F_rh=F_rh, V_2=V_2, alpha_xx=alpha_xx)
    return result


def calculate_wave_resistance(V_2, h_0, g, T, L, B, C_P, C_WP, lcb, L_R, A_T, C_M, delta, rho):
    """
    Calculate the wave resistance and related hydrodynamic coefficients for a ship.

    Parameters
    ----------
    V_2v : float
        Karpov corrected ship speed relative to water (m/s).
    h_0 : float
        Water depth (m).
    g : float
        Gravitational acceleration (m/s^2).
    T : float
        Ship's draft (m).
    L : float
        Ship's length at waterline (m).
    B : float
        Ship's beam at waterline (m).
    C_P : float
        Prismatic coefficient (dimensionless).
    C_WP : float
        Waterplane area coefficient (dimensionless).
    lcb : float
        Longitudinal center of buoyancy (as a fraction of L, dimensionless).
    L_R : float
        Length of run (m).
    A_T : float
        Transom area (m^2).
    C_M : float
        Midship section coefficient (dimensionless).
    delta : float
        Displacement volume (m^3).
    rho : float
        Water density (kg/m^3).

    Returns
    -------
    F_rL : float
        Froude number based on ship's speed and length.
    i_E : float
        Half angle of entrance (degrees).
    c_1 : float
        Empirical coefficient for wave resistance.
    c_2 : float
        Bulbous bow effect coefficient.
    c_5 : float
        Transom stern influence coefficient.
    c_7 : float
        Coefficient based on B/L ratio.
    c_15 : float
        Coefficient based on L^3/delta ratio.
    c_16 : float
        Coefficient based on prismatic coefficient.
    lmbda : float
        Lambda parameter for wave resistance calculation.
    m_1 : float
        Exponential coefficient for wave resistance.
    m_2 : float
        Cosine coefficient for wave resistance.
    R_W : float
        Calculated wave resistance (kN).

    """
    # checks
    assert g >= 0, f"g should be positive: {g}"
    assert L >= 0, f"L should be positive: {L}"

    F_rL = V_2 / np.sqrt(g * L)  # Froude number based on ship's speed to water and its length of waterline

    # parameter c_7 is determined by the B/L ratio
    # Van Koningsveld et al (2023) - Part IV Table 5.1
    if B / L < 0.11:
        c_7 = 0.229577 * (B / L) ** 0.33333
    if B / L > 0.25:
        c_7 = 0.5 - 0.0625 * (L / B)
    else:
        c_7 = B / L

    # half angle of entrance in degrees
    # Van Koningsveld et al (2023) - Part IV Table 5.1
    i_E = 1 + 89 * np.exp(
        -((L / B) ** 0.80856)
        * ((1 - C_WP) ** 0.30484)
        * ((1 - C_P - 0.0225 * lcb) ** 0.6367)
        * ((L_R / B) ** 0.34574)
        * ((100 * delta / (L**3)) ** 0.16302)
    )

    # Van Koningsveld et al (2023) - Part IV Table 5.1
    c_1 = 2223105 * (c_7**3.78613) * ((T / B) ** 1.07961) * (90 - i_E) ** (-1.37165)
    # TODO: check if we need to adapt this, for cases where we do want to calculate with bulbous bows
    c_2 = 1  # accounts for the effect of the bulbous bow, which is not present at inland ships
    c_5 = 1 - (0.8 * A_T) / (B * T * C_M)  # influence of the transom stern on the wave resistance

    # parameter c_15 depends on the ratio L^3 / delta
    # Van Koningsveld et al (2023) - Part IV Table 5.1
    if (L**3) / delta < 512:
        c_15 = -1.69385
    elif (L**3) / delta > 1727:
        c_15 = 0
    else:
        c_15 = -1.69385 + (L / (delta ** (1 / 3)) - 8) / 2.36

    # parameter c_16 depends on C_P
    # Van Koningsveld et al (2023) - Part IV Table 5.1
    if C_P < 0.8:
        c_16 = 8.07981 * C_P - 13.8673 * (C_P**2) + 6.984388 * (C_P**3)
    else:
        c_16 = 1.73014 - 0.7067 * C_P

    if L / B < 12:
        lmbda = 1.446 * C_P - 0.03 * (L / B)
    else:
        lmbda = 1.446 * C_P - 0.36

    # Van Koningsveld et al (2023) - Part IV Table 5.1
    m_1 = 0.0140407 * (L / T) - 1.75254 * ((delta) ** (1 / 3) / L) - 4.79323 * (B / L) - c_16
    # Van Koningsveld et al (2023) - Part IV Table 5.1
    m_2 = c_15 * (C_P**2) * np.exp((-0.1) * (F_rL ** (-2)))

    # Van Koningsveld et al (2023) - Part IV Eq 5.16
    # Segers (2019) distinguishes multiple Froude classes (Section 3.2.5 and Appendix C - C.2).
    # for all reasonable combinations of ship lengths and speeds, inland ships always fall in the
    # F_n,V_2 < 0.4 class
    R_W = c_1 * c_2 * c_5 * delta * rho * g * np.exp(m_1 * (F_rL ** (-0.9)) + m_2 * np.cos(lmbda * (F_rL ** (-2)))) / 1000  # kN

    wave_resistance_properties = collections.namedtuple(
        "WaveResistanceProperties",
        [
            "F_rL",
            "i_E",
            "c_1",
            "c_2",
            "c_5",
            "c_7",
            "c_15",
            "c_16",
            "lmbda",
            "m_1",
            "m_2",
            "R_W",
        ],
    )
    result = wave_resistance_properties(
        F_rL=F_rL,
        i_E=i_E,
        c_1=c_1,
        c_2=c_2,
        c_5=c_5,  
        c_7=c_7,
        c_15=c_15,
        c_16=c_16,
        lmbda=lmbda,
        m_1=m_1,
        m_2=m_2,
        R_W=R_W,
    )   
    return result


def calculate_residual_resistance(V_2, g, A_T, B, C_WP, rho, T, L, C_B, S, T_F, h_B, A_BT, bulbous_bow):
    """
    Calculate the residual resistance of a ship, which includes the resistance due to
    the immersed transom, model-ship correlation resistance, and bulbous bow resistance.

    This function computes the residual resistance components based on the ship's
    speed, transom area, beam, waterplane coefficient, density, draft, length,
    block coefficient, wetted surface area, and bulbous bow presence.

    Parameters
    ----------
    V_2 : float
        The corrected velocity (m/s) according to the Karpov method.
    g : float
        Gravitational acceleration (m/s^2).
    A_T : float
        Traverse area of the transom (m^2). Van Koningsveld et al (2023) - below Eq 5.16
    B : float
        Beam of the ship (m).
    C_WP : float
        Waterplane coefficient (dimensionless).
    rho : float
        Density of the water (kg/m^3).
    T : float
        Actual draft of the ship (m).
    L : float
        Length of the ship (m).
    C_B : float
        Block coefficient of the ship (dimensionless).
    S : float
        Wetted surface area of the ship (m^2).
    T_F : float
        Draft at the forefoot (m).
    h_B : float
        Height of the bulbous bow (m).
    A_BT : float
        Area of the bulbous bow transom (m^2).
    bulbous_bow : bool
        Whether the ship has a bulbous bow (True) or not (False).

    Returns
    -------
    F_nT : float
        Froude number based on transom immersion (dimensionless).
    c_6 : float
        Coefficient for resistance due to immersed transom (dimensionless).
    R_TR : float
        Resistance due to immersed transom (kN).
    c_4 : float
        Coefficient for model-ship correlation resistance (dimensionless).
    c_2 : float
        Coefficient for model-ship correlation resistance (dimensionless).
    C_A : float
        Model-ship correlation resistance coefficient (dimensionless).
    R_A : float
        Model-ship correlation resistance (kN).
    F_ni : float
        Froude number based on immersion of bulbous bow (dimensionless).
    P_B : float
        Coefficient for the emergence of bulbous bow (dimensionless).
    R_B : float
        Resistance due to the bulbous bow (kN).
    R_res : float
        Total residual resistance (kN).

    """

    # Resistance due to immersed transom: R_TR [kN]
    F_nT = V_2 / np.sqrt(2 * g * A_T / (B + B * C_WP))  # Froude number based on transom immersion
    assert not isinstance(F_nT, complex), f"residual? froude number should not be complex: {F_nT}"

    c_6 = 0.2 * (1 - 0.2 * F_nT)  # Assuming F_nT < 5, this is the expression for coefficient c_6

    R_TR = (0.5 * rho * (V_2**2) * A_T * c_6) / 1000

    # Model-ship correlation resistance: R_A [kN]

    if T / L < 0.04:
        c_4 = T / L
    else:
        c_4 = 0.04
    c_2 = 1

    C_A = 0.006 * (L + 100) ** (-0.16) - 0.00205 + 0.003 * np.sqrt(L / 7.5) * (C_B**4) * c_2 * (0.04 - c_4)
    assert not isinstance(C_A, complex), f"C_A number should not be complex: {C_A}"

    R_A = (0.5 * rho * (V_2**2) * S * C_A) / 1000  # kW

    # Resistance due to the bulbous bow (R_B)

    # Froude number based on immersoin of bulbous bow [-]
    F_ni = V_2 / np.sqrt(g * (T_F - h_B - 0.25 * np.sqrt(A_BT) + 0.15 * V_2**2))

    P_B = (0.56 * np.sqrt(A_BT)) / (T_F - 1.5 * h_B)  # P_B is coefficient for the emergence of bulbous bow
    if bulbous_bow:
        R_B = ((0.11 * np.exp(-3 * P_B**2) * F_ni**3 * A_BT**1.5 * rho * g) / (1 + F_ni**2)) / 1000
    else:
        R_B = 0

    R_res = R_TR + R_A + R_B

    residual_resistance_properties = collections.namedtuple(
        "ResidualResistanceProperties",
        [
            "F_nT",
            "c_6",
            "R_TR",
            "c_4",
            "c_2",
            "C_A",
            "R_A",
            "F_ni",
            "P_B",
            "R_B",
            "R_res",
        ],
    )
    result = residual_resistance_properties(
        F_nT=F_nT,  
        c_6=c_6,
        R_TR=R_TR,
        c_4=c_4,
        c_2=c_2,
        C_A=C_A,
        R_A=R_A,
        F_ni=F_ni,
        P_B=P_B,
        R_B=R_B,
        R_res=R_res,
    )       
    return result


def calculate_total_resistance(v, g, h_0, C_B, L, B, T, bulbous_bow, C_BB, nu, rho, c_stern, one_k2):
    """
    Calculate the total resistance of a ship, which includes frictional, viscous,
    appendage, wave, and residual resistance components.

    The total resistance
    R_tot [kN] = R_f * (1+k1) + R_APP + R_W + R_TR + R_A

    Parameters
    ----------
    v : float
        Ship's speed relative to water (m/s).
    g : float
        Gravitational acceleration (m/s^2).
    h_0 : float
        Water depth (m).
    C_B : float
        Block coefficient of the ship (dimensionless).
    L : float
        Length of the ship (m).
    B : float
        Beam of the ship (m).
    T : float
        Actual draft of the ship (m).
    bulbous_bow : bool
        Whether the ship has a bulbous bow (True) or not (False).
    C_BB : float
        Beam coefficient of bulbous bow (dimensionless).
    nu : float
        Kinematic viscosity of water (m^2/s).
    rho : float
        Density of water (kg/m^3).
    c_stern : float
        Determines the shape of the afterbody (dimensionless).
    one_k2 : float
        Appendage resistance factor (1 + k2) (dimensionless).

    Returns
    -------
    R_tot : float
        Total resistance of the ship (kN).
    """

    # TODO: this function is rather odd as it calls all other resistance functions,
    #  computing lots of unused values (which are set in the corresponding method),
    #  hence the function is not yet used in the main class method.

    # vessel properties
    C_M, C_WP, C_P, delta, lcb, L_R, A_T, A_BT, S, S_APP, S_B, T_F, h_B = calculate_properties(C_B, L, B, T, bulbous_bow, C_BB)

    # frictional resistance
    R_f, C_f, _, _, _, _, _, _, _, _ = calculate_frictional_resistance(v, h_0, L, nu, T, S, S_B, rho)

    # viscous resistance
    _, one_k1, _ = calculate_viscous_resistance(c_stern, B, L, T, L_R, C_P, R_f, delta)

    # appendage resistance
    R_APP = calculate_appendage_resistance(v, rho, S_APP, one_k2, C_f)

    # calculate the Karpov corrected velocity
    F_rh, V_2, alpha_xx = karpov(v, h_0, g, T)
    print('Original v = {} m/s, Karpov corrected V_2 = {} m/s'.format(v, V_2))

    # wave resistance
    _, _, _, _, _, _, _, _, _, _, _, R_W = calculate_wave_resistance(V_2, h_0, g, T, L, B, C_P, C_WP, lcb, L_R, A_T, C_M, delta, rho)

    # residual resistance
    _, _, R_TR, _, _, _, R_A, _, _, R_B, _ = calculate_residual_resistance(
        V_2, g, A_T, B, C_WP, rho, T, L, C_B, S, T_F, h_B, A_BT, bulbous_bow
    )

    # The total resistance R_tot [kN] = R_f * (1+k1) + R_APP + R_W + R_TR + R_A
    R_tot = R_f * one_k1 + R_APP + R_W + R_TR + R_A + R_B

    return R_tot


def calculate_total_power_required(
    v,
    h_0,
    R_tot,
    F_rL,
    x,
    C_B,
    delta,
    D_s,
    eta_o,
    eta_r,
    eta_t,
    eta_g,
    P_hotel,
    P_installed,
):
    """
    Calculate the total power required for a ship based on its speed, resistance,
    and various efficiency factors.

    Parameters
    ----------
    v : float
        Ship's speed relative to water (m/s).
    h_0 : float
        Water depth (m).
    R_tot : float
        Total resistance of the ship (kN).
    F_rL : float
        Froude number based on ship's speed and length (dimensionless).
    x : float
        Number of propellers (dimensionless).
    C_B : float
        Block coefficient of the ship (dimensionless).
    delta : float
        Displacement volume of the ship (m^3).
    D_s : float
        Propeller diameter (m).
    eta_o : float
        Overall efficiency of the propulsion system (dimensionless).
    eta_r : float
        Efficiency of the reduction gear (dimensionless).
    eta_t : float
        Efficiency of the transmission (dimensionless).
    eta_g : float
        Efficiency of the generator (dimensionless).
    P_hotel : float
        Hotel load power (kW).
    P_installed : float
        Installed power of the ship (kW).

    Returns
    -------
    P_e : float
        Required power for propulsion (kW).
    dw : float
        Velocity correction coefficient (dimensionless).
    w : float
        Wake fraction (dimensionless).
    t : float
        Thrust deduction factor (dimensionless).
    eta_h : float
        Hull efficiency (dimensionless).
    P_d : float,
        Power delivered (kW)
    P_b : float
        Power at the brake (kW)
    P_propulsion : float
        Power required for propulsion (kW).
    P_tot : float
        Total power required for the ship (propulsion + hotel power) (kW).
    P_given : float
        Power given by the captain (kW).
    P_partial : float
        Partial engine load (P_tot / P_installed) (dimensionless).

    Notes
    -----
    The P_b calculation involves gearing efficiency and transmission efficiency
    already, while P_d does not. P_d is the power delivered to propellers. To estimate
    the renewable fuel use, 'energy conversion efficiencies' are taken into consideration.

    The 'energy conversion efficiencies' for renewable fuel powered vessels are
    commonly measured/given as a whole covering the engine power systems, including
    different engines (such as fuel cell engine, battery engine, internal
    combustion engine, hybrid engine) efficiencies, and corresponding gearbox
    efficiencies, AC/DC converter efficiencies, excluding the efficiency items of
    propellers.

    It is therefore, important carefully align with the later use of 'energy conversion
    efficiencies' for fuel use estimation and to prevent double use of some power efficiencies,
    such as gearing efficiency. It is important to carefully consider if it is best to use
    P_d or P_b as power starting point.

    """

    # Effective Horse Power (EHP), 'P_e' - power associated with the vessel's speed and its resistance
    # (Van Koningsveld et al (2023) - Part IV Eq 5.17)
    P_e = v * R_tot

    # velocity correction coefficient, 'dw'
    # Segers (2021) (http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3)
    # Appendix C
    if F_rL < 0.2:
        dw = 0  # the velocity correction coefficient is 0 when FrL is smaller than 0.2
    else:
        dw = 0.1  # otherwise the velocity correction coefficient is 0.1

    # wake fraction, 'w'
    # Segers (2021) (http://resolver.tudelft.nl/uuid:a260bc48-c6ce-4f7c-b14a-e681d2e528e3)
    # Appendix C - Eq C.1
    w = 0.11 * (0.16 / x) * C_B * np.sqrt((delta ** (1 / 3)) / D_s) - dw  # wake fraction 'w'

    assert not isinstance(w, complex), f"w should not be complex: {w}"

    # thrust deduction factor, 't'
    if x == 1:
        # if the ship has 1 propeller
        # (Van Koningsveld et al (2023) - Part IV Eq 5.22)
        t = 0.6 * w * (1 + 0.67 * w)
    else:
        # (Van Koningsveld et al (2023) - Part IV Eq 5.23)
        t = 0.8 * w * (1 + 0.25 * w)

    # hull efficiency 'eta_h'
    eta_h = (1 - t) / (1 - w)  # hull efficiency eta_h

    # TODO: check below suggestions.
    #   They were made to allow for better translation to alternative energy sources.
    #   But the changes induced unexpected behaviour.

    # Calculation hydrodynamic efficiency eta_D  according to Simic et al (2013) "On Energy Efficiency of Inland
    # Waterway Self-Propelled Cargo Vessels", https://www.researchgate.net/publication/269103117
    # hydrodynamic efficiency eta_D is a ratio of power used to propel the ship and delivered power

    # relation between eta_D and ship velocity v
    # if h_0 >= 9:
    #     if F_rh >= 0.5:
    #         eta_D = 0.6
    #     elif 0.325 <= F_rh < 0.5:
    #         eta_D = 0.7
    #     elif 0.28 <= F_rh < 0.325:
    #         eta_D = 0.59
    #     elif 0.2 < F_rh < 0.28:
    #         eta_D = 0.56
    #     elif 0.17 < F_rh <= 0.2:
    #         eta_D = 0.41
    #     elif 0.15 < F_rh <= 0.17:
    #         eta_D = 0.35
    #     else:
    #         eta_D = 0.29
    #
    # elif 5 <= h_0 < 9:
    #     if F_rh > 0.62:
    #         eta_D = 0.7
    #     elif 0.58 < F_rh < 0.62:
    #         eta_D = 0.68
    #     elif 0.57 < F_rh <= 0.58:
    #         eta_D = 0.7
    #     elif 0.51 < F_rh <= 0.57:
    #         eta_D = 0.68
    #     elif 0.475 < F_rh <= 0.51:
    #         eta_D = 0.53
    #     elif 0.45 < F_rh <= 0.475:
    #         eta_D = 0.4
    #     elif 0.36 < F_rh <= 0.45:
    #         eta_D = 0.37
    #     elif 0.33 < F_rh <= 0.36:
    #         eta_D = 0.36
    #     elif 0.3 < F_rh <= 0.33:
    #         eta_D = 0.35
    #     elif 0.28 < F_rh <= 0.3:
    #         eta_D = 0.331
    #     else:
    #         eta_D = 0.33
    # else:
    #     if F_rh > 0.56:
    #         eta_D = 0.28
    #     elif 0.4 < F_rh <= 0.56:
    #         eta_D = 0.275
    #     elif 0.36 < F_rh <= 0.4:
    #         eta_D = 0.345
    #     elif 0.33 < F_rh <= 0.36:
    #         eta_D = 0.28
    #     elif 0.3 < F_rh <= 0.33:
    #         eta_D = 0.27
    #     elif 0.28 < F_rh <= 0.3:
    #         eta_D = 0.26
    #     else:
    #         eta_D = 0.25
    #
    # # Delivered Horse Power (DHP), P_d
    # P_d = P_e / eta_D

    # logger.debug("eta_D = {:.2f}".format(eta_D))

    # Delivered Horse Power (DHP), 'P_d'
    # (Van Koningsveld et al (2023) - Part IV Eq 5.19)
    P_d = P_e / (eta_o * eta_r * eta_h)

    # Brake Horse Power (BHP), 'P_b'
    # (Van Koningsveld et al (2023) - Part IV Eq 5.24)
    P_b = P_d / (eta_t * eta_g)

    # TODO: consider how to integrate the suggestion to use P_propulsion = P_d
    #  When working with alternative energy carriers it may be better to use
    #  Delivered horse power, the power delivered to propellers, rather than
    #  Power at the Brake. This may avoid double counting of efficiencies.

    # Propulsion Power, 'P_propulsion'
    P_propulsion = P_b  # propulsion power is defined here as Power at the Brake.

    # Total Power, 'P_tot'
    P_tot = P_hotel + P_propulsion

    # Partial engine load, 'P_partial': used in the 'Emission calculations'
    if P_tot > P_installed:
        P_given = P_installed
        P_partial = 1
    else:
        P_given = P_tot
        P_partial = P_tot / P_installed

    logger.debug(f"The total power required is {P_tot} kW")
    logger.debug(f"The actual total power given is {P_given} kW")
    logger.debug(f"The partial load is {P_partial}")

    assert not isinstance(P_given, complex), f"P_given number should not be complex: {P_given}"

    total_power_required_properties = collections.namedtuple(
        "TotalPowerRequiredProperties",
        [
            "P_e",
            "dw",
            "w",
            "t",
            "eta_h",
            "P_d",
            "P_b",
            "P_propulsion",
            "P_tot",
            "P_given",
            "P_partial",
        ],
    )
    result = total_power_required_properties(
        P_e=P_e,
        dw=dw,
        w=w,
        t=t,
        eta_h=eta_h,
        P_d=P_d,
        P_b=P_b,
        P_propulsion=P_propulsion,
        P_tot=P_tot,
        P_given=P_given,
        P_partial=P_partial,
    )   
    return result



"""
Mixin classes for movable objects.

The following classes are provided:
- Movable
- ContainerDependentMovable
"""
# package(s) for documentation, debugging, saving and loading
import logging
import warnings
import deprecated
from typing import Union

# math packages
import numpy as np

# spatial libraries
import pyproj
import shapely
import shapely.geometry
from shapely import Geometry
import networkx as nx
import simpy

# time packages
import datetime

# use OpenCLSim objects for core objects (identifiable is imported for later use)
import opentnsim.strategy
from openclsim.core import SimpyObject, Locatable, Log
from opentnsim.core.container import HasContainer
from opentnsim.energy.mixins import ConsumesEnergy
from opentnsim.graph.mixins import get_length_of_edge

# get logger
logger = logging.getLogger(__name__)


class Routable(SimpyObject):
    """Mixin class: Something with a route (networkx format)

    Parameters
    ----------
    route: list
        list of node-IDs
    complete_path: list, optional
        ???
    args, kwargs:
        passed to SimpyObject. Must at least contain parameter env: simpy.Environment.

    Attributes
    -----------
    route: list
        list of node-IDs
    complete_path: list, optional
        ???
    position_on_route: int
        index of position on the route
    """

    def __init__(self, route, complete_path=None, *args, **kwargs):
        """Initialization"""
        # check env input
        env = kwargs.get("env")
        # if env is given and env is not None
        if env is not None:
            has_fg = hasattr(env, "FG")
            has_graph = hasattr(env, "graph")
            if has_fg and not has_graph:
                warnings.warn(".FG attribute has been renamed to .graph, please update your code", DeprecationWarning)
                env.graph = env.FG
            assert (
                has_fg or has_graph
            ), "Routable expects `.graph` (a networkx graph) to be present as an attribute on the environment"

        # initialization
        super().__init__(*args, **kwargs)
        self.route = route
        # start at start of route
        self.position_on_route = 0
        self.complete_path = complete_path

        self.check_attributes()

    def check_attributes(self):
        """Check if all required attributes are set."""
        # check if route is set
        if self.route is None or not isinstance(self.route, list):
            raise ValueError("Routable requires a route (list of node IDs) to be set")
        # check if env is set
        if not hasattr(self, "env") or not isinstance(self.env, simpy.Environment):
            raise ValueError("Routable requires an environment (simpy.Environment) to be set")
        # check if route is on graph
        if not all(node in self.graph.nodes for node in self.route):
            raise ValueError("Routable route must be on the graph")

    @property
    def graph(self):
        """
        Return the graph of the underlying environment.

        If it's multigraph cast to corresponding type
        If you want the multidigraph use the HasMultiGraph mixin

        """
        graph = None
        if hasattr(self.env, "graph"):
            graph = self.env.graph
        else:
            raise ValueError("Routable expects .graph to be present on env")

        if isinstance(graph, nx.MultiDiGraph):
            return nx.DiGraph(graph)
        elif isinstance(graph, nx.MultiGraph):
            return nx.Graph(graph)
        return graph


@deprecated.deprecated(reason="Use Routable instead of Routeable")
class Routeable(Routable):
    """Old name for Mixin class: renamed to Routable."""


class Movable(Locatable, Routable, Log):
    """Mixin class: Something can move.

    Used for object that can move with a fixed speed

    Parameters
    ----------
    v: float
        speed of the object (in m/s)
    geometry: shapely.geometry.Point
        passed to Locatable. point used to track its current location
    node: str, optional
        passed to Locatable,
    route: list, optional
        passed to Routable,
    complete_path: list, optional
        passed to Routable,

    Attributes
    ----------
    v: float
        speed of the object (in m/s)
    on_pass_edge_functions: list
        list of functions to call when passing an edge
    on_pass_node_functions: list
        list of functions to call when passing a node
    wsg84: pyproj.Geod
        used for distance computation
    req: simpy.Resource request
        resource request for passing nodes and edges. saved for using resources over various nodes and edges.
    resource: simpy.Resource
        resource used for passing nodes and edges. saved for using resources over various nodes and edges.

    Notes
    -----
    This class can handle a basic graph in env.graph. This will result in a simple simulation.

    The class can also handle more complex simulations. For this, extra information is needed in the graph, such as:
    - Resources on nodes and edges, which can be requested and released.
        - Resources on nodes are saved in env.graph.nodes[node]["Resources"]
        - Resources on edges are saved in env.graph.edges[origin, destination]["Resources"].
        - Several edges and nodes can have the same resource, which is usefull when a segment can only be used by one vessel at a time.
        - When using a digraph, make sure to assign the same resource to both directions of the edge.
    - Current on edges, which can be used to compute the speed of the vessel.
        - Current on edges is saved in env.graph.edges[origin, destination]["Info"]["Current"].
        - Current can only be used in a directed graph (DiGraph).
        - Current is positive in the direction of the edge, and negative in the opposite direction.
        - Make sure to assign current to both directions of the edge in a digraph. (the negative and positive current)
    - Power information, which can be used to compute the speed of the vessel.
        - self must be a mixin of ConsumesEnergy.
        - self must have the attribute P_tot_given and must not be None.
        - general depth of fairway is saved in env.graph.edges[origin, destination]["Info"]["GeneralDepth"].
        - width of fairway is saved in env.graph.edges[origin, destination]["Info"]["Width"]. If not given, we use 150 m.

    """

    def __init__(self, v: float, *args, **kwargs):
        """Initialization"""
        super().__init__(*args, **kwargs)
        self.v = v
        self.distance = 0
        self.on_pass_node_functions = []
        self.on_pass_edge_functions = []
        self.on_complete_pass_edge_functions = []
        self.on_look_ahead_to_node_functions = []
        self.wgs84 = pyproj.Geod(ellps="WGS84")

        self._check_attributes()

        # resource memory for passing nodes and edges
        self.req = None
        self.resource = None

        # keep track of distance travelled on edge
        self.distance_left_on_edge = np.nan

    def _check_attributes(self):
        """Check if all required attributes are set."""
        # each node on route should have a geometry
        geoms = nx.get_node_attributes(self.graph, "geometry")
        if not all(node in geoms for node in self.route):
            raise ValueError(
                "Nodes on route must have a geometry attribute. Missing geometries for nodes: {}".format(
                    [node for node in self.route if node not in geoms]
                )
            )

    @property
    def current_node(self) -> Union[str, None]:
        """Return the current node on the route based on self.position_on_route."""
        if 0 <= self.position_on_route < len(self.route):
            return self.route[self.position_on_route]
        else:
            return None

    @property
    def next_node(self) -> Union[str, None]:
        """Return the next node on the route based on self.position_on_route."""
        if 0 <= self.position_on_route < len(self.route) - 1:
            return self.route[self.position_on_route + 1]
        else:
            return None

    @property
    def route_ahead(self):
        """Return the remaining route ahead of the current position."""
        if 0 <= self.position_on_route < len(self.route):
            return self.route[self.position_on_route :]
        else:
            return []

    def determine_route_to_target_node(self, target_node: str):
        """Determine the route to the target node.
        Parameters
        ----------
        target_node: str
            The target node to determine the route to.
        Returns
        -------
        list
            The route to the target node.
        """
        if target_node not in self.route_ahead:
            raise ValueError("Target node must be in the remaining route ahead.")
        # get index of first occurrence of target_node in route_ahead
        try:
            idx = self.route_ahead.index(target_node)
        except ValueError:
            warnings.warn(f"No route found to waiting area")
            return []

        # get route to target node
        route = self.route_ahead[: idx + 1]

        return route

    def update_position(self, position_on_route: int):
        """Update the position on the route.

        Parameters
        ----------
        position_on_route: int
            index of position on the route

        """
        self.position_on_route = position_on_route
        self.geometry = nx.get_node_attributes(self.graph, "geometry")[self.current_node]

    # TODO: Move was eerst een functie met 'destination' als argument, maar dat is nu niet meer het geval. Willen we dat dit weg is?
    def move(self):
        """Moves vessel over the path defined by self.route.

        Assumption is that self.path is in the right order - vessel moves from route[0] to route[-1].

        Yields
        ------
        time it takes to travel the distance to the destination.

        """

        # Check if vessel has arrival time and let vessel wait to start moving
        if hasattr(self, "metadata") and "arrival_time" in self.metadata:
            arrival_time = self.metadata['arrival_time']
            current_time = datetime.datetime.fromtimestamp(self.env.now)
            delay = (arrival_time - current_time).total_seconds()
            yield self.env.timeout(delay)

        # Check if vessel is at correct location - if not, move to location
        yield from self._move_to_start()

        # look ahead to first node
        self.position_on_route = 0
        yield from self.look_ahead_to_node(self.route[0])

        # Move over the path and log every step
        for index, edge in enumerate(zip(self.route[:-1], self.route[1:])):
            # update current position
            self.update_position(index)

            yield from self.pass_node(self.current_node)

            # are we already at destination?
            if self.next_node == self.current_node:
                warnings.warn(
                    "Route passes node {} twice consecutively..".format(self.current_node),
                    UserWarning,
                )
                continue

            yield from self.pass_edge(self.current_node, self.next_node)
            yield from self.complete_pass_edge(self.next_node)

            # we arrived at destination
            # update to new position
            self.update_position(index + 1)

            # look ahead to next node
            yield from self.look_ahead_to_node(self.current_node)

        # arrived at end of route. release resource if needed
        if self.req is not None:
            self._release_resource()

        logger.debug("  distance: " + "%4.2f" % self.distance + " m")
        if self.current_speed is not None:
            logger.debug("  sailing:  " + "%4.2f" % self.current_speed + " m/s")
            logger.debug("  duration: " + "%4.2f" % ((self.distance / self.current_speed) / 3600) + " hrs")
        else:
            logger.debug("  current_speed:  not set")

    def _move_to_start(self):
        """Move vessel to the start of the route.

        Yields
        ------
        The time it takes to move to the start of the route.
        """
        # Check if vessel is at correct location - if not, move to location
        vessel_origin_location = nx.get_node_attributes(self.env.graph, "geometry")[self.route[0]]
        if self.geometry != vessel_origin_location:
            self.log_entry_v0("Sailing to start start", self.env.now, self.distance, self.geometry)
            start_location = self.geometry
            logger.debug("Origin: {orig}")
            logger.debug("Destination: {dest}")

            self.distance += self.wgs84.inv(start_location.x, start_location.y, vessel_origin_location.x, vessel_origin_location.y)[
                2
            ]

            yield self.env.timeout(self.distance / self.current_speed)
            self.geometry = vessel_origin_location
            self.log_entry_v0("Sailing to start stop", self.env.now, self.distance, self.geometry)

    def pass_node(self, node):
        """pass a node and call all on_pass_node_functions

        Parameters
        ----------
        node: str
            the node to pass

        Yields
        ------
        The time it takes to pass the node.
        """
        # request resource if needed
        if "Resources" in self.graph.nodes[node].keys() and self.req is None:
            arrival = self.env.now  # remember when we arrived at the node
            # request
            yield from self._request_resource(self.graph.nodes[node]["Resources"])

            # we had to wait, log it
            if arrival != self.env.now:
                self.log_entry_v0(
                    "Waiting to pass node {} start".format(node),
                    arrival,
                    self.distance,
                    self.graph.nodes[node]["geometry"],
                )
                self.log_entry_v0(
                    "Waiting to pass node {} stop".format(node),
                    self.env.now,
                    self.distance,
                    self.graph.nodes[node]["geometry"],
                )

        # call all on_pass_node_functions
        for on_pass_node_function in self.on_pass_node_functions:
            yield from on_pass_node_function(node)

        # release resource if needed
        if self.req is not None:
            # only release if resource is not needed in the next edge
            if self.next_edge is None:
                self._release_resource()
            if "Resources" not in self.graph.edges[self.next_edge]:
                self._release_resource()
            elif self.graph.edges[self.next_edge]["Resources"] != self.resource:
                self._release_resource()
            else:
                pass

    def _release_resource(self):
        """Release the current resource."""
        self.resource.release(self.req)
        self.req = None
        self.resource = None

    def _request_resource(self, resource):
        """Request a resource for passing nodes and edges.

        Parameters
        ----------
        resource: simpy.Resource
            the resource to request
        Yields
        ------
        simpy.Resource request
        """
        self.resource = resource
        self.req = self.resource.request()
        yield self.req

    @property
    def next_edge(self):
        """Return the next edge on the route. based on self.position_on_route.

        Returns
        -------
        tuple(str, str) or None
            (origin, destination) of the next edge on the route.
        """
        if self.position_on_route < len(self.route) - 1:
            return self.route[self.position_on_route], self.route[self.position_on_route + 1]
        else:
            return None

    def pass_edge(self, origin, destination):
        """pass an edge and call all on_pass_edge_functions.

        Parameters
        ----------
        origin: str
            the origin node of the edge
        destination: str
            the destination node of the edge

        Yields
        ------
        The time it takes to pass the edge.
        """
        edge = (origin, destination)
        edge_info = self.graph.edges[edge]
        orig = nx.get_node_attributes(self.graph, "geometry")[origin]
        dest = nx.get_node_attributes(self.graph, "geometry")[destination]
        distance = get_length_of_edge(self.graph, edge)
        self.distance_left_on_edge = distance

        # calculate velocity based on depth and power, if possible.
        self.v = self._compute_velocity_on_edge(origin, destination)

        # Check if the edge has current info
        # NB: positive current is directed from origin to destination
        current = self._get_current(origin, destination)
        # Wait for edge resources to become available
        # TODO: Misschien moeten we Resources ook onder Info hangen?
        if "Resources" in edge_info.keys() and self.req is None:
            arrival = self.env.now  # remember when we arrived at the edge
            yield from self._request_resource(self.graph.edges[origin, destination]["Resources"])
            # we had to wait, log it
            if arrival != self.env.now:
                self.log_entry_v0(
                    "Waiting to pass edge {} - {} start".format(origin, destination),
                    arrival,
                    self.distance,
                    orig,
                )
                self.log_entry_v0(
                    "Waiting to pass edge {} - {} stop".format(origin, destination),
                    self.env.now,
                    self.distance,
                    orig,
                )

        self.log_entry_v0(
            "Sailing from node {} to node {} start".format(self.current_node, self.next_node),
            self.env.now,
            self.distance,
            orig,
        )

        # on pass edge functions
        for on_pass_edge_function in self.on_pass_edge_functions:
            yield from on_pass_edge_function(origin, destination)

        # default velocity based on current speed.
        timeout = self.distance_left_on_edge / (self.current_speed + current)
        yield self.env.timeout(timeout)
        self.distance += self.distance_left_on_edge

        self.log_entry_v0(
            "Sailing from node {} to node {} stop".format(self.current_node, self.next_node),
            self.env.now,
            self.distance,  # TODO distance klopt nu  niet na een sluismodule
            dest,
        )
        self.geometry = dest

        # release resource if needed
        if "Resources" in edge_info.keys():
            # only release if resource is not needed in the next node
            if "Resources" not in self.graph.nodes[destination].keys():
                self._release_resource()
            elif self.resource != self.graph.nodes[destination]["Resources"]:
                self._release_resource()
                self.resource = None
            else:
                pass

    def complete_pass_edge(self, destination):
        for gen in self.on_complete_pass_edge_functions:
            yield from gen(destination)

    def look_ahead_to_node(self, destination):

        for gen in self.on_look_ahead_to_node_functions:
            yield from gen(destination)

    def _get_current(self, origin, destination):
        """Get the current on the edge

        Parameters
        ----------
        origin: str
            the origin node of the edge
        destination: str
            the destination node of the edge

        Returns
        -------
        float
            the current on the edge (in m/s)
        """
        if "Info" not in self.graph.edges[origin, destination].keys():
            # no info on the current, return 0
            return 0.0
        elif "Current" not in self.graph.edges[origin, destination]["Info"].keys():
            # no info on current, return 0
            return 0.0
        elif not isinstance(self.graph, nx.DiGraph):
            raise TypeError(
                "Current is only available on a DiGraph. Use a Digraph to use current in your calculations.",
                UserWarning,
            )
            return 0.0
        current = self.graph.edges[origin, destination]["Info"]["Current"]

        if (self.current_speed + current) <= 0:
            raise ValueError(
                f"Current {current} m/s is larger than current speed {self.current_speed} m/s. "
                "This will result in a negative speed, which is not allowed.",
                UserWarning,
            )
        return current

    @property
    def current_speed(self):
        """return the current speed of the vessel"""
        return self.v

    def _compute_velocity_on_edge(self, origin, destination):
        """compute the velocity on an edge, based on the energy module and the depth.

        parameters
        ----------
        origin: str
            the origin node of the edge
        destination: str
            the destination node of the edge
        """

        edge = (origin, destination)
        if hasattr(self,'overruled_speed') and edge in self.overruled_speed.index:
            overruled_speed = self.overruled_speed.loc[edge]
            return overruled_speed

        # check if we have the energy mixin and ptot_given
        if not isinstance(self, ConsumesEnergy):
            return self.v
        elif self.P_tot_given is None:
            return self.v

        # determine the depth of the edge
        edge = self.graph.edges[origin, destination]
        try:
            depth = edge["Info"]["GeneralDepth"]
        except KeyError:
            raise ValueError(
                f"Edge {origin} - {destination} has no GeneralDepth in Info. " f"\n Add info or remove ConsumesEnergy mixin"
            )

        # You can input more power than is realistic
        # There are two mechanisms that reduce the power given:
        # 1. The grounding speed:
        # TODO: Als we dit laten staan, moeten we get_upperbound_for_power2v ook checken en testen.
        # TODO get_upperbound_for_power2v heeft een width standaard 150. Is dat handig?
        edge_width = self._get_general_width(origin, destination)
        edge_width = edge_width if edge_width is not None else 150  # default width if not set

        upperbound = opentnsim.strategy.get_upperbound_for_power2v_optim(self, width=edge_width, depth=depth, margin=0)
        # Here the upperbound is used to estimate the actual velocity
        power_used = min(self.P_tot_given, upperbound)
        return self.power2v(self, edge, power_used)

    def _get_general_width(self, origin, destination):
        """Get the general width of the edge.

        Parameters
        ----------
        origin: str
            the origin node of the edge
        destination: str
            the destination node of the edge

        Returns
        -------
        float
            the general width of the edge (in m)
        """
        if "Info" not in self.graph.edges[origin, destination].keys():
            return None
        elif "GeneralWidth" not in self.graph.edges[origin, destination]["Info"].keys():
            return None
        else:
            return self.graph.edges[origin, destination]["Info"]["GeneralWidth"]


class ContainerDependentMovable(Movable, HasContainer):
    """Mixin class: ContainerDependentMovable class
    Used for objects that move with a speed dependent on the container level

    Parameters
    ----------
    compute_v: function
        a function, given the fraction the container is filled (in [0,1]), returns the current speed
    v: float
        passed to Movable, speed of the object (in m/s)
    geometry: shapely.geometry.Point
        passed to Movable. point used to track its current location
    node: str, optional
        passed to Movable,
    route: list, optional
        passed to Movable,
    complete_path: list, optional
        passed to Movable,
    Capacity: float
        passed to HasContainer, the capacity of the container, which may either be continuous (like water) or discrete (like apples)
    level: int, default=0
        passed to HasContainer, level of the container at the beginning of the simulation
    total_requested: int, default=0
        passed to HasContainer, total amount that has been requested at the beginning of the simulation

    Attributes
    ----------
    compute_v: function
        a function, given the fraction the container is filled (in [0,1]), returns the current speed
    current_speed: float
        the current speed of the vessel (in m/s), based on the filling degree of the container
    """

    def __init__(self, compute_v, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """Initialization"""
        self.compute_v = compute_v
        self.wgs84 = pyproj.Geod(ellps="WGS84")

    @property
    def current_speed(self):
        """return the current speed of the vessel, based on the filling degree of the container"""
        return self.compute_v(self.filling_degree)



"""
Core utiltities related to logging.
"""

# %% IMPORT DEPENDENCIES
# generic
import pandas as pd

from shapely import Point

# internal
import opentnsim.graph.mixins as mixins


# # %% CONVERT LOG TO EVENT TABLE
# def logbook2eventtable(objs):
#     """
#     Transform object logbooks into a 'minimum event table'.

#     Implements the basic 'event table' concept as proposed by Van der Werff:

#     Van der Werff, S.E., F. Baart and M. van Koningsveld (2025). “Merging Multiple
#     System Perspectives: The Key to Effective Inland Shipping Emission-Reduction
#     Policy Design.” Journal of Marine Science and engineering 13(4), 716.
#     https://doi.org/10.3390/jmse13040716

#     Van der Werff, S.E., S. Eppenga, A. van der Hout, F. Baart and M. van
#     Koningsveld (2025). “Multi-perspective nautical safety risk assessment of
#     allisions with offshore wind parks.” Applied Ocean Research 158(2025),104564.
#     https://doi.org/10.1016/j.apor.2025.104564

#     For waterborne traffic over a network, a unique event is defined by:
#      - a unique vessel,
#      - a specific section of the waterway,
#      - a specific time.

#     Parameters
#     ----------
#     objs: list
#         List of OpenTNSim simulation objects with log information.

#     Returns
#     -------
#     eventtable: pandas.DataFrame
#         DataFrame with all events from obj.logbook attributes in objs.
#     """
#     # check if all objects have a logbook with expected structure
#     for obj in objs:
#         if (
#             not hasattr(obj, "logbook")
#             or not hasattr(obj, "id")
#             or not hasattr(obj, "name")
#         ):
#             raise ValueError(
#                 f"Object {obj} does not have a logbook or id/name attributes."
#             )

#     # construct all logged events
#     events = []
#     for obj in objs:
#         df = pd.DataFrame.from_dict(obj.logbook)

#         unique_activities = df["Message"].str.replace(" start", "").str.replace(" stop", "").unique()
#         for activity in unique_activities:
#             start_time = df[df.Message == activity + " start"]["Timestamp"].values[0]
#             stop_time = df[df.Message == activity + " stop"]["Timestamp"].values[0]

#             start_location = df[df.Message == activity + " start"]["Geometry"].values[0]
#             stop_location = df[df.Message == activity + " stop"]["Geometry"].values[0]

#             duration_seconds = (stop_time - start_time) / pd.Timedelta(seconds=1)

#             if isinstance(start_location, Point):
#                 distance_meters = opentnsim.graph.calculate_distance(start_location, stop_location)
#             else:
#                 distance_meters = None

#             events.append(
#                 {
#                     "object id": obj.id,
#                     "object name": obj.name,
#                     "activity name": activity,
#                     "start location": start_location,
#                     "stop location": stop_location,
#                     "start time": start_time,
#                     "stop time": stop_time,
#                     "distance (m)": distance_meters,
#                     "duration (s)": duration_seconds,
#                 }
#             )

#     # Final DataFrame
#     eventtable = pd.DataFrame(events)

#     return eventtable

def logbook2eventtable(objs):
    """
    Transform object logbooks into a 'minimum event table'.

    Implements the basic 'event table' concept as proposed by Van der Werff:

    Van der Werff, S.E., F. Baart and M. van Koningsveld (2025). “Merging Multiple
    System Perspectives: The Key to Effective Inland Shipping Emission-Reduction
    Policy Design.” Journal of Marine Science and engineering 13(4), 716.
    https://doi.org/10.3390/jmse13040716

    Van der Werff, S.E., S. Eppenga, A. van der Hout, F. Baart and M. van
    Koningsveld (2025). “Multi-perspective nautical safety risk assessment of
    allisions with offshore wind parks.” Applied Ocean Research 158(2025),104564.
    https://doi.org/10.1016/j.apor.2025.104564

    For waterborne traffic over a network, a unique event is defined by:
     - a unique vessel,
     - a specific section of the waterway,
     - a specific time.

    Parameters
    ----------
    objs: list
        List of OpenTNSim simulation objects with log information.

    Returns
    -------
    eventtable: pandas.DataFrame
        DataFrame with all events from obj.logbook attributes in objs.
    """
    # check if all objects have a logbook with expected structure
    for obj in objs:
        if (
            not hasattr(obj, "logbook")
            or not hasattr(obj, "id")
            or not hasattr(obj, "name")
        ):
            raise ValueError(
                f"Object {obj} does not have a logbook or id/name attributes."
            )

    # construct all logged events
    events = []
    for obj in objs:
        df = pd.DataFrame.from_dict(obj.logbook)
        df.sort_values(by="Timestamp", inplace=True)

        for i in range(0, len(df)):
            start_row = df.iloc[i]
            if start_row["Message"].endswith(" start"):
                activity = start_row["Message"].replace(" start", "")
            else:
                continue  # skip non-start messages

            stop_row = df[(df["Message"] == activity + " stop") & (df["Timestamp"] > df.iloc[i]["Timestamp"])].iloc[0]

            start_time = start_row["Timestamp"]
            stop_time = stop_row["Timestamp"]
            start_location = start_row["Geometry"]
            stop_location = stop_row["Geometry"]

            duration_seconds = (stop_time - start_time).total_seconds()
            if isinstance(start_location, Point):
                distance_meters = mixins.calculate_distance(start_location, stop_location)
            else:
                distance_meters = None

            events.append(
                {
                    "object id": obj.id,
                    "object name": obj.name,
                    "activity name": activity,
                    "start location": start_location,
                    "stop location": stop_location,
                    "start time": start_time,
                    "stop time": stop_time,
                    "distance (m)": distance_meters,
                    "duration (s)": duration_seconds,
                }
            )

    # Final DataFrame
    eventtable = pd.DataFrame(events)

    return eventtable



# -*- coding: utf-8 -*-

"""Graph module."""
import functools
# packkage(s) for documentation, debugging, saving and loading
import logging
import os
import pickle
import uuid
from itertools import cycle

# matplotlib
import matplotlib.pyplot as plt
# package(s) for data handling
import networkx as nx
import yaml
import numpy as np
import plotly.graph_objects as go
# spatial libraries
import pyproj
import requests
import shapely.geometry
# package(s) related to the simulation
import simpy
from opentnsim.core import Identifiable, Locatable
# OpenTNSim
from opentnsim.graph import mixins as graph_module
from opentnsim.graph import utils
from plotly.offline import init_notebook_mode, iplot
from shapely.geometry import LineString, Point
from shapely.ops import transform

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# Determine the wgs84 geoid
wgs84 = pyproj.Geod(ellps="WGS84")


def get_geometry_of_edge(graph, edge):
    """get the geometry of the edge in WGS84

    Parameters
    ----------
    graph: networkx.Graph
        The graph object.
    edge : tuple
        The edge to get the length of. is a tuple of two node-names.

    Returns
    -------
    float
        The length of the edge in meters.
    """

    edge_info = graph.edges[edge]
    if "geometry" not in edge_info:
        orig = nx.get_node_attributes(graph, "geometry")[edge[0]]
        dest = nx.get_node_attributes(graph, "geometry")[edge[1]]
        geometry = LineString([orig,dest])
        graph.edges[edge]["geometry"] = geometry
    else:
        geometry = graph.edges[edge]["geometry"]

    coordinates_x = geometry.coords.xy[0]
    coordinates_y = geometry.coords.xy[1]
    min_coordinates_x = np.min(coordinates_x)
    max_coordinates_x = np.max(coordinates_x)
    min_coordinates_y = np.min(coordinates_y)
    max_coordinates_y = np.max(coordinates_y)
    if not isinstance(edge_info["geometry"], shapely.geometry.LineString):
        raise ValueError(f"Edge geometry in edge {edge}: attribute must be a shapely LineString.")
    if min_coordinates_x < -180. or max_coordinates_x > 180. or min_coordinates_y < -90. or max_coordinates_y > 90.:
        raise ValueError(f"Edge geometry in edge {edge}: attribute is not defined in WGS84.")

    return geometry


def determine_length_of_edge_geometry(graph, edge, current_crs="EPSG:4326", crs_meter="EPSG:4087"):
    wgs84 = pyproj.CRS(current_crs)
    wgs84_m = pyproj.CRS(crs_meter)
    wgs84_to_wgs84_m = pyproj.transformer.Transformer.from_crs(wgs84, wgs84_m, always_xy=True).transform
    geometry = get_geometry_of_edge(graph, edge)
    geometry_m = transform(wgs84_to_wgs84_m, geometry)
    length_m = geometry_m.length
    return length_m


def get_length_of_edge(graph, edge, current_crs="EPSG:4326", crs_meter="EPSG:4087"):
    """get the length of an edge in meters

    Parameters
    ----------
    graph: networkx.Graph
        The graph object.
    edge : tuple
        The edge to get the length of. is a tuple of two node-names.

    Returns
    -------
    float
        The length of the edge in meters.
    """

    edge_info = graph.edges[edge]
    if "length_m" in edge_info:
        pass
    else:
        length_m = determine_length_of_edge_geometry(graph, edge, current_crs, crs_meter)
        graph.edges[edge]["length_m"] = length_m

    return edge_info["length_m"]


def find_closest_node(G, point):
    """find the closest node on the graph from a given point"""

    distance = np.full((len(G.nodes)), fill_value=np.nan)
    for ii, n in enumerate(G.nodes):
        distance[ii] = point.distance(G.nodes[n]["geometry"])
    name_node = list(G.nodes)[np.argmin(distance)]
    distance_node = np.min(distance)

    return name_node, distance_node


def calculate_distance(geom_start, geom_stop):
    """method to calculate the distance (as the bird flies) in meters between two geometries

    Parameters
    ----------
    geom_start : shapely.geometry.Point
        Starting point geometry. must contain x and y attributes.
    geom_stop : shapely.geometry.Point
        Stopping point geometry. must contain x and y attributes.

    Returns
    -------
    float
        Distance in meters between the two geometries.
    """

    wgs84 = pyproj.Geod(ellps="WGS84")

    # distance between two points
    return float(wgs84.inv(geom_start.x, geom_start.y, geom_stop.x, geom_stop.y)[2])


def calculate_distance_along_path(graph, path):
    """method to calculate the greater circle distance along path in meters from WGS84 coordinates

    Parameters
    ----------
    graph : networkx.Graph
        The graph object.
    path : list
        List of nodes that together form a path.

    Returns
    -------
    float
        Path length in meters.
    """

    path_length = 0

    for node in enumerate(path[:-1]):
        orig = nx.get_node_attributes(graph, "geometry")[path[node[0]]]
        dest = nx.get_node_attributes(graph, "geometry")[path[node[0]+1]]
        path_length += calculate_distance(orig, dest)

        if node[0] + 2 == len(path):
                    break

    return path_length

def calculate_depth(geom_start, geom_stop, graph):
    """method to calculate the depth of the waterway in meters between two geometries.

    Parameters
    ----------
    geom_start : shapely.geometry.Point
        Starting point geometry. Must represent a node in graph graph.
    geom_stop : shapely.geometry.Point
        Stopping point geometry. must represent a node in graph graph.
    graph : networkx.Graph
        The graph containing vaarweginformatie.nl data, with nodes and edges.
        Must contain 'Info' attribute on edges with 'GeneralDepth'.
        Must contain an edge between geom_start and geom_stop.

    Returns
    -------
    float
        The depth of the waterway between the two geometries in meters.

    Raises
    ------
    ValueError
        If geom_start or geom_stop are not nodes in the graph graph.
        If there is no edge between the two nodes in the graph graph.
        If the depth data is not available for the edge between the two nodes.
    """

    depth = 0

    # The node on the graph of vaarweginformatie.nl closest to geom_start and geom_stop

    node_start = find_closest_node(graph, geom_start)[0]
    node_stop = find_closest_node(graph, geom_stop)[0]

    # Read from the graph data from vaarweginformatie.nl the General depth of each edge
    # TODO: check it this needs to be made more general, now relies on ['Info'] to be present
    if node_start == node_stop:
        return np.nan  # if the start and stop nodes are the same, return 0 depth

    try:
        if "Info" in graph.get_edge_data(node_start, node_stop).keys():
            depth = graph.get_edge_data(node_start, node_stop)["Info"]["GeneralDepth"]

        elif "GeneralDepth" in graph.get_edge_data(node_start, node_stop).keys():
            depth = graph.get_edge_data(node_start, node_stop)["GeneralDepth"]
        else:
            return np.nan  # if no depth data is available, return NaN
    except:
        depth = np.nan  # When there is no data of the depth available of this edge, it gives a message

    h_0 = depth

    # depth of waterway between two points
    return h_0


def geom_to_edges(geom, properties):
    """Generate edges from a geometry, yielding an edge id and edge properties. The edge_id consists of a tuple of coordinates"""
    if geom.geom_type not in ["LineString", "MultiLineString"]:
        msg = "Only ['LineString', 'MultiLineString'] are supported, got {}".format(geom.geom_type)
        raise ValueError(msg)
    if geom.geom_type == "MultiLineString":
        for geom in geom.geoms:
            yield from geom_to_edges(geom, properties)
    elif geom.geom_type == "LineString":
        edge_properties = properties.copy()
        edge_source_coord = geom.coords[0]
        edge_target_coord = geom.coords[-1]
        edge_properties["Wkt"] = shapely.wkt.dumps(geom)
        edge_properties["Wkb"] = shapely.wkb.dumps(geom)
        edge_properties["Json"] = shapely.geometry.mapping(geom)
        edge_properties["e"] = [edge_source_coord, edge_target_coord]
        edge_id = (edge_source_coord, edge_target_coord)
        yield edge_id, edge_properties


def geom_to_node(geom: shapely.geometry.Point, properties: dict):
    if not geom.geom_type == "Point":
        msg = "Only 'Point' is supported, got {}".format(geom.geom_type)
        raise ValueError(msg)
    node_properties = properties.copy()
    node_properties["Wkt"] = shapely.wkt.dumps(geom)
    node_properties["Wkb"] = shapely.wkb.dumps(geom)
    node_properties["Json"] = shapely.geometry.mapping(geom)
    node_properties["n"] = geom.coords[0]
    node_id = geom.coords[0]
    return node_id, node_properties


def gdf_to_nx(gdf):
    """Convert a geopandas dataframe to a networkx DiGraph"""
    graph = nx.DiGraph()
    for _, feature in gdf.iterrows():
        geom = feature.geometry
        if geom is None:
            raise nx.NetworkXError("Bad data: feature missing geometry")
        properties = feature.drop(labels=["geometry"])
        # in case we have single points in the geometry, add them as nodes
        if geom.geom_type == "Point":
            node_idx = geom.coords[0]
            graph.add_node(node_idx, **properties)
            continue
        if geom.geom_type in ["LineString", "MultiLineString"]:
            for edge_id, edge_properties in geom_to_edges(geom, properties):
                node_source, node_target = edge_properties["e"]
                source_geom = shapely.geometry.Point(*node_source)
                _, node_properties = geom_to_node(source_geom, {})
                graph.add_node(edge_id[0], **node_properties)
                _, node_properties = geom_to_node(source_geom, {})
                graph.add_node(edge_id[1], **node_properties)
                graph.add_edge(edge_id[0], edge_id[1], **edge_properties)
    return graph


class Node(Identifiable, Locatable):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DiGraph:

    def __init__(self, edges, weights=[1], geometries=[None], edges_info={}, crs="EPSG:4326", bidirectional=True, *args, **kwargs):
        """edges: a list of tuples of two Node-objects"""

        super().__init__(*args, **kwargs)
        self.graph = nx.DiGraph()
        CRS = pyproj.CRS(crs)
        wgs84 = pyproj.CRS("EPSG:4326")
        CRS_to_wgs84 = pyproj.Transformer.from_crs(CRS, wgs84, always_xy=True).transform
        for index, ((node_I, node_II), weight, geometry, edge_info) in enumerate(
            zip(edges, cycle(weights), cycle(geometries), cycle([edges_info]))
        ):
            if node_I.name not in self.graph.nodes:
                node_I.geometry = transform(CRS_to_wgs84, node_I.geometry)
                self.graph.add_node(node_I.name, geometry=node_I.geometry)
            if node_II.name not in self.graph.nodes:
                node_II.geometry = transform(CRS_to_wgs84, node_II.geometry)
                self.graph.add_node(node_II.name, geometry=node_II.geometry)
            if not geometry:
                geometry = LineString([node_I.geometry, node_II.geometry])
            geod = pyproj.Geod(ellps="WGS84")
            length = geod.geometry_length(geometry)
            Info = {}
            for key, value in edge_info.items():
                Info[key] = value[index]
            self.graph.add_edge(
                node_I.name,
                node_II.name,
                weight=weight,
                geometry=geometry,
                length=length,
                Info=Info,
            )
            if bidirectional:
                self.graph.add_edge(
                    node_II.name,
                    node_I.name,
                    weight=weight,
                    geometry=geometry.reverse(),
                    length=length,
                    Info=Info,
                )


class Graph:
    """General networkx object

    Initialize a nx.Graph() element

    Attributes
    ----------
    graph : networkx.Graph
        The graph object
    graph_info : dict
        The graph information
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graph = nx.Graph()
        self.graph_info = utils.info(self.graph)

    def from_shape(self, file_location, shapefile, simplify=True, strict=True):
        """Generate nx.Graph() from shapefile
        Make sure to install the required package gdal.

        run pip show gdal to check if gdal is installed.

        Parameters
        ----------
        file_location: Path
            location on server of the shapefile
        shapefile: str
            name of the shapefile (including .shp)
        simplify: bool
            if True, the graph is simplified
        strict: bool
            if True, the graph is strict
        """
        from osgeo import ogr, osr

        # Create graph
        self.graph = opentnsim.utils.read_shp(os.path.join(file_location, shapefile), simplify=simplify, strict=strict)
        self.graph_info = opentnsim.utils.info(self.graph)

        # Get spatial reference
        driver = ogr.GetDriverByName("ESRI Shapefile")
        dataset = driver.Open(os.path.join(file_location, shapefile))
        self.SpatialRef = dataset.GetLayer().GetSpatialRef()

    def transform_projection(self, to_EPSG):
        """create a transformation object to transform the graph to a new projection
        Make sure to install the required package gdal.

        run pip show gdal to check if gdal is installed.
        Parameters
        ----------
        to_EPSG: int
            The EPSG code to transform the graph to
        """

        from osgeo import ogr, osr

        outSpatialRef = osr.SpatialReference()
        outSpatialRef.ImportFromEPSG(to_EPSG)

        # Transform the coordinates
        transform = osr.CoordinateTransformation(self.SpatialRef, outSpatialRef)

        return transform

    def change_projection(self, transform, point):
        """Transform one point on the graph

        Make sure to install the required package gdal (for osgeo).
        run pip show gdal to check if gdal is installed.

        Parameters
        ----------
        transform:
        """
        from osgeo import ogr, osr

        point = ogr.CreateGeometryFromWkt(str(point))

        point.Transform(transform)
        point.ExportToWkt()

        return point.GetX(), point.GetY()

    def create_graph_new_projection(self, to_EPSG=4326):
        """redefine self.graph with the new projection

        Make sure to install the required package gdal (for osgeo).
        run pip show gdal to check if gdal is installed.

        Parameters
        ----------
        to_EPSG: int
            The EPSG code to transform the graph to
        """
        new_graph = nx.Graph()
        transform = self.transform_projection(to_EPSG)

        # Required to prevent loop-in-loop
        nodes_dict = {}

        # Add original nodes and edges to new graph
        for i, node in enumerate(self.graph.nodes(data=True)):
            # TODO: depending on the coordinate transformation x, y might refer to x,y or latitude, longitude.
            # Shapely assumes always x/lon, y/lat
            coordinates = self.change_projection(
                transform,
                shapely.geometry.Point(list(self.graph.nodes)[i][0], list(self.graph.nodes)[i][1]),
            )
            name = "({:f}, {:f})".format(coordinates[1], coordinates[0])
            geometry = shapely.geometry.Point(coordinates[1], coordinates[0])

            nodes_dict[list(self.graph.nodes)[i]] = name
            new_graph.add_node(name, name=name, Position=(coordinates[1], coordinates[0]), geometry=geometry, Old=node[1])

        for edge in self.graph.edges(data=True):
            node_1 = nodes_dict[edge[0]]
            node_2 = nodes_dict[edge[1]]

            new_graph.add_edge(node_1, node_2, Info=edge[2])

        new_graph = new_graph.to_directed()

        if opentnsim.utils.info(new_graph) != self.graph_info:
            self.graph = new_graph
            self.graph_info = opentnsim.utils.info(new_graph)
        else:
            print("Conversion did not create an exact similar graph")

            print("")
            print("Original graph")
            print(self.graph_info)

            print("")
            print("New graph")
            print(opentnsim.utils.info(new_graph))

            self.graph = new_graph
            self.graph_info = opentnsim.utils.info(new_graph)

    def add_resources(self, edges, resources, environment):
        """Add resources to the edges of the graph

        Parameters
        ----------
        edges: list
            List of edges to which the resources should be added
        resources: list
            List of resources to be added to the edges. Should be same length as edges
        environment: simpy.Environment
            The simpy environment to which the resources should be added
        """
        for i, edge in enumerate(edges):
            self.graph.edges[edge]["Resources"] = simpy.Resource(environment, capacity=resources[i])

    def plot(
        self,
        size=[10, 10],
        with_labels=False,
        node_size=0.5,
        font_size=2,
        width=0.2,
        arrowsize=3,
    ):
        """Plot the graph
        Parameters
        ----------
        size: list
            The size of the figure
        with_labels: bool
            If True, the labels of the nodes are shown
        node_size: int
            The size of the nodes, default is 0.5
        font_size: int
            The size of the font, default is 2
        width: int
            The width of the edges, default is 0.2
        arrowsize: int
            The size of the arrows, default is 3
        """
        plt.figure(figsize=size)

        # If graph has positional attributes
        try:
            nx.draw(
                self.graph,
                nx.get_node_attributes(self.graph, "Position"),
                with_labels=with_labels,
                node_size=node_size,
                font_size=font_size,
                width=width,
                arrowsize=arrowsize,
            )
        # If graph does not have any positional information
        except:
            nx.draw(self.graph)

        plt.show()


def get_minimum_depth(graph, route):
    """return the minimum depth on the route based on the GeneralDepth in the Info dictionary

    Parameters
    ----------
    graph: networkx.Graph
        The graph object. Edges in the graph should have a property called Info (dict), with key GeneralDepth
    route: list
        The route to check the depth for. The route is a list of node ids.

    Returns
    -------
    float
        The minimum depth on the route
    """
    # loop over the route
    depths = []
    # loop over all node pairs (e: edge numbers)
    for e in zip(route[:-1], route[1:]):
        # get the properties
        edge = graph.get_edge_data(e[0], e[1])
        # lookup the depth
        depth = edge["Info"]["GeneralDepth"]
        # remember
        depths.append(depth)
        # find the minimum
    h_min = np.min(depths)
    return h_min


def compute_distance(edge, orig, dest):
    """compute distance from origin to destination.
    The distance is computed based on the edge geometry.
    If the edge has no geometry, returns the distance 'as the crow flies'.

    Parameters
    ----------
    edge: dict
        The edge to compute the distance for.
    orig: shapely.geometry.Point
        The origin point
    dest: shapely.geometry.Point
        The destination point

    """
    if "geometry" not in edge:
        distance = wgs84.inv(
            shapely.geometry.shape(orig).x,
            shapely.geometry.shape(orig).y,
            shapely.geometry.shape(dest).x,
            shapely.geometry.shape(dest).y,
        )[2]
        return distance

    edge_route = np.array(edge["geometry"].coords)

    # check if edge is in the sailing direction, otherwise flip it
    distance_from_start = wgs84.inv(
        orig.x,
        orig.y,
        edge_route[0][0],
        edge_route[0][1],
    )[2]
    distance_from_stop = wgs84.inv(
        orig.x,
        orig.y,
        edge_route[-1][0],
        edge_route[-1][1],
    )[2]
    if distance_from_start > distance_from_stop:
        # when the distance from the starting point is greater than from the end point
        edge_route = np.flipud(np.array(edge["geometry"].coords))

    distance = 0
    for index, pt in enumerate(edge_route[:-1]):
        sub_orig = shapely.geometry.Point(edge_route[index][0], edge_route[index][1])
        sub_dest = shapely.geometry.Point(edge_route[index + 1][0], edge_route[index + 1][1])

        distance += wgs84.inv(
            shapely.geometry.asShape(sub_orig).x,
            shapely.geometry.asShape(sub_orig).y,
            shapely.geometry.asShape(sub_dest).x,
            shapely.geometry.asShape(sub_dest).y,
        )[2]
    return distance


class FIS:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @functools.lru_cache
    def load_fis_network(url):
        """load the topological fairway information system network (vaarweginformatie.nl)"""

        # get the data from the url
        resp = requests.get(url)
        # convert to file object
        stream = io.StringIO(resp.text)

        # This will take a minute or two
        # Here we convert the network to a networkx object
        G = yaml.load(stream, Loader=yaml.Loader)

        # some brief info
        n_bytes = len(resp.content)
        msg = """Loaded network from {url} file size {mb:.2f}MB. Network has {n_nodes} nodes and {n_edges} edges."""
        summary = msg.format(url=url, mb=n_bytes / 1000**2, n_edges=len(G.edges), n_nodes=len(G.nodes))
        logger.info(summary)

        # The topological network contains information about the original geometry.
        # Let's convert those into python shapely objects for easier use later
        for n in G.nodes:
            G.nodes[n]["geometry"] = shapely.geometry.Point(G.nodes[n]["X"], G.nodes[n]["Y"])
        for e in G.edges:
            edge = G.edges[e]
            edge["geometry"] = shapely.wkt.loads(edge["Wkt"])

        return G

    @staticmethod
    def import_FIS(url):

        fname = "fis_cache\\{}.pkl".format("FIS")
        if os.path.exists(fname):
            print("I am loading cached network")
            with open(fname, "rb") as pkl_file:
                graph = pickle.load(pkl_file)
                pkl_file.close()

        else:
            print("I am getting new network")
            graph = FIS.load_fis_network(url)

            os.makedirs(os.path.dirname(fname), exist_ok=True)
            with open(fname, "wb") as pkl_file:
                pickle.dump(graph, pkl_file)
                pkl_file.close()

        return graph


class HasMultiDiGraph:
    """This locking module uses a MultiDiGraph to represent the network. This converts other graphs to a MultiDiGraph."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def multidigraph(self):
        # create a multidigraph copy of graph if it was not done before
        if hasattr(self,"env"):
            graph_class = self.env
        else:
            graph_class = self
        if not hasattr(graph_class, "_multidigraph"):
            graph_class._multidigraph = self.copy()
        return graph_class._multidigraph

    def copy(self):
        if hasattr(self,"env"):
            graph_class = self.env
        else:
            graph_class = self
        multidigraph = graph_class.graph
        if not isinstance(graph_class.graph, nx.MultiDiGraph):
            multidigraph = nx.MultiDiGraph(multidigraph)
        return multidigraph


def plot_graph(graph, static: bool = False):
    """method to plot a graph

    Parameters
    ----------
    graph : networkx.Graph
        A graph object.
    static : bool, optional
        If True, returns a static Plotly figure object.
        If False, displays the figure

    Returns
    -------
    fig : plotly.graph_objs._figure.Figure
        Object that contains a graph figure.
    """

    # Labels
    labels = {node: node for node in graph.nodes()}
    edge_labels = {(u, v): f"{d['weight']} km" for u, v, d in graph.edges(data=True)}

    # positions
    positions = {node: (graph.nodes[node]["geometry"].x, graph.nodes[node]["geometry"].y) for node in graph.nodes}

    # Edge labels in meters
    edge_labels = {}
    for u, v in graph.edges():
        origin = graph.nodes[u]['geometry']
        destination = graph.nodes[v]['geometry']
        distance_m = graph_module.calculate_distance(origin, destination)
        edge_labels[(u, v)] = f"{int(distance_m)} m"

    # Edge traces and arrow annotations
    edge_traces = []
    arrow_annotations = []
    for u, v in graph.edges():
        x0, y0 = positions[u]
        x1, y1 = positions[v]
        edge_traces.append(go.Scatter(
            x=[x0, x1],
            y=[y0, y1],
            line=dict(width=2, color='red'),
            mode='lines',
            hoverinfo='none'
        ))
        arrow_annotations.append(go.layout.Annotation(
            x=x1, y=y1,
            ax=x0, ay=y0,
            xref='x', yref='y',
            axref='x', ayref='y',
            showarrow=True,
            arrowhead=2,  # Closed arrowhead
            arrowsize=2,
            arrowwidth=1,
            arrowcolor='red'
        ))

    # Node trace
    node_trace = go.Scatter(
        x=[positions[node][0] for node in graph.nodes()],
        y=[positions[node][1] for node in graph.nodes()],
        mode='markers+text',
        marker=dict(color='darkblue', size=20),
        text=[labels[node] for node in graph.nodes()],
        textposition='middle center',
        textfont=dict(color='white', size=15),
        hoverinfo='text'
    )

    # Edge label annotations
    edge_label_annotations = []
    for (u, v), label in edge_labels.items():
        x0, y0 = positions[u]
        x1, y1 = positions[v]
        x_mid = (x0 + x1) / 2
        y_mid = (y0 + y1) / 2
        edge_label_annotations.append(go.layout.Annotation(
            x=x_mid, y=y_mid,
            text=label,
            showarrow=False,
            font=dict(color='black', size=15)
        ))

    # Combine annotations
    annotations = arrow_annotations + edge_label_annotations

    # Create figure
    fig = go.Figure(data=edge_traces + [node_trace])
    fig.update_layout(
        title="Directed Geographic Network Graph (WGS84 Projection) with Edge Lengths in Meters",
        xaxis=dict(title="Longitude", showgrid=False, zeroline=False, showticklabels=False),
        yaxis=dict(title="Latitude", showgrid=False, zeroline=False, showticklabels=False),
        annotations=annotations,
        showlegend=False,
        plot_bgcolor='white',
    )

    if static is False:
        # Initialize notebook mode for Plotly
        init_notebook_mode(connected=True)
        # Display the figure in a Jupyter notebook
        iplot(fig)
    else:
        return fig
